/*
 * SPDX-FileCopyrightText: 2024-2025 Bruno Anselme <be.root@gfree.fr>
 *
 * Authors:
 *   Bruno Anselme <be.root@gfree.fr>
 *
 *   SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This file is part of the NoComprendo software.
*/

#ifndef CONFIG_H
#define CONFIG_H

#include <QDir>
#include <QFile>
#include <QUrl>
#include <QPoint>
#include <QSize>
#include <QMap>

#define APPCONF Config::instance()

// This is a singleton class
// Use it like this: Config::instance().var_or_function or with macro: APPCONF.var_or_function

class Config {
public:
    // Constructors
    Config();
    Config(Config const&) = delete;             // disallow clone constructor
    void operator=(Config const&)  = delete;    // disallow copy constructor
    // Static instance
    static Config& instance() {
        static Config    instance_;             // Guaranteed to be destroyed. Instantiated on first use.
        return instance_;
    }
    void        trace();                        // Trace function for debugging

public:
    QString     lang;                   //!< Language (en_US, fr_FR...)
    QString     model;                  //!< Vosk language model
    // Globals paths
    QDir        appPath;                //!< Application path
    QDir        voskSystemPath;         //!< Vosk models system path
    QDir        voskUserPath;           //!< Vosk models user path
    QDir        userSetsPath;           //!< User commands sets path
    QString     setsDirName;            //!< Short dir name
    // Vosk server url
    QUrl        voskUrl;                //!< Path to language models to download
    QUrl        voskJson;               //!< Json file name with models informations
    // Files names (full paths)
    QString     mouseFile;              //!< Utterances file for Mouse MoveTo
    QString     repeaterFile;           //!< Utterances file for meta Repeat
    // General settings
    bool        autoStart;              //!< Auto start speech recognition
    bool        removeBadUtterance;     //!< Remove bad utterance flag
    bool        useSysTray;             //!< Use a system tray icon
    int         rFontSize;              //!< Reminder font size
    bool        rInvertColor;           //!< Inverted colors for reminder
    bool        ask2Quit;               //!< Confirm before exit
    int         maxlines;               //!< Max line count in console
    bool        startReduced;           //!< Main window starting state
    // OSD settings
    bool        osdOn;                  //!< OSD on/off flag
    bool        osdAll;                 //!< Display all in OSD
    int         osdDuration;            //!< OSD duration (milliseconds)
    QPoint      osdPosition;            //!< OSD position
    QSize       osdSize;                //!< OSD size
    // Keyboard settings
    QString     keyboardVariant;        //!< Keyboard variant for dotool
    int         keyDelay;               //!< Delay between keys
    int         keyHold;                //!< Key hold delay
    int         typeDelay;              //!< Delay between chars
    int         typeHold;               //!< Char hold delay
    // Mouse settings
    int         mouseStep;              //!< Mouse step for MoveLeft, MoveRight, MoveUp, MoveDown commands
    int         glideStep;              //!< Mouse glide step
    int         wheelStep;              //!< Mouse wheel step
    int         glideRate;              //!< Mouse glide rate (milliseconds)
    bool        numPadGrid;             //!< Mouse grid natural order or numpad order
    // Audio settings
    QByteArray  audioId;                //!< QAudioDevice id for microphone
    // Dictation
    bool        transparentMode;        //!< Transparent or opaque
    bool        osdNoDictation;         //!< Don't show dictated text with osd
    bool        capitalizeFirst;        //!< Dictation capitalize first letter
    QString     dictationSuffix;        //!< String added after each dictation
    // Text to speech
    QDir        speechDir;              //!< Directory containing avalaible pico2wave languages
    QString     speechLang;             //!< Default language for speech
    QMap<QString, QString> speechesList;
    // Themes
    int         themeIndex;
    // Command line options
    bool        commandTrace;           //!< Print utterances to stdout
};

#endif // CONFIG_H
