/*
 * SPDX-FileCopyrightText: 2024-2025 Bruno Anselme <be.root@gfree.fr>
 *
 * Authors:
 *   Bruno Anselme <be.root@gfree.fr>
 *
 *   SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This file is part of the NoComprendo software.
*/

#include <iostream>
#include <QFile>
#include <QDebug>
#include <QLocale>
#include <QThread>
#include <QMessageBox>
#include <QApplication>

#include "app_version.h"
#include "dotoolpipe.h"
#include "misc.h"

#define DOTOOL_XKB_LAYOUT "DOTOOL_XKB_LAYOUT"       // Keyboard environment variable for dotool
#define DOTOOL_XKB_VARIANT "DOTOOL_XKB_VARIANT"     // Keyboard variant environment variable for dotool

static QHash<QString, QString> const  qtToKeysym = {
    {"Ctrl",         "ctrl"}
  , {"Shift",        "shift"}
  , {"Alt",          "alt"}
  , {"Meta",         "super"}

  , {"-",            "minus"}
  , {"=",            "equal"}
  , {"PgUp",         "pageup"}
  , {"PgDown",       "pagedown"}
  , {"Backspace",    "backspace"}
  , {"Del",          "delete"}
  , {"Ins",          "insert"}
  , {"Return",       "enter"}

  , {";",            "semicolon"}
  , {"'",            "apostrophe"}
  , {"`",            "grave"}
  , {"\\",           "backslash"}
  , {",",            "comma"}
  , {".",            "dot"}
  , {"/",            "slash"}
  , {"Ins",          "insert"}
  , {"Del",          "delete"}

  , {"Print",        "print"}
  , {"ScrollLock",   "scrolllock"}
  , {"NumLock",      "numlock"}
  , {"CapsLock",     "capslock"}
  , {"Pause",        "pause"}

  , {"Num++",        "kpplus"}
  , {"Num+-",        "kpminus"}
  , {"Num+/",        "kpslash"}
  , {"Num+.",        "kpdot"}

// Wrong behavior with function keys. Use keycode.
  , {"F1",           "k:59"}
  , {"F2",           "k:60"}
  , {"F3",           "k:61"}
  , {"F4",           "k:62"}
  , {"F5",           "k:63"}
  , {"F6",           "k:64"}
  , {"F7",           "k:65"}
  , {"F8",           "k:66"}
  , {"F9",           "k:67"}
  , {"F10",          "k:68"}
  , {"F11",          "k:87"}
  , {"F12",          "k:88"}

// Media keys (don't work)
  , {"Volume Mute",      "mute"}
  , {"Volume Up",        "volumeup"}
  , {"Volume Down",      "volumedown"}
  , {"Media Previous",   "previoussong"}
  , {"Media Next",       "nextsong"}
  , {"Media Play",       "playpause"}
  , {"Media Stop",       "stop"}
};

DoToolPipe::DoToolPipe(QString &kbdVariant)
    : QObject()
    , program()        // OS dotool. Fail with alt+f6, alt+f7 (no action)
    , myProcess()
    , moveStep(10)
    , glideStep(2)
    , wheelStep(5)
    , errorCount(0)
{
    // Set DOTOOL_XKB_LAYOUT environment variable, mandatory for dotool
    if (QLocale::system().name().left(2).toUtf8() != "en")
        if (!qputenv(DOTOOL_XKB_LAYOUT,  QLocale::system().name().sliced(0, 2).toUtf8()))      // Set keyboard locale for dotool. Should not fail.
            qWarning() << "DOTOOL_XKB_LAYOUT couldn't be assigned";

    // Set DOTOOL_XKB_VARIANT environment variable if provided
    if (!qputenv(DOTOOL_XKB_VARIANT, kbdVariant.toUtf8()))      // Set keyboard variant for dotool. Should not fail.
        qWarning() << "DOTOOL_XKB_VARIANT couldn't be assigned";

    // qWarning() << "dotool started with" << kbdVariant;

    program = findProgram("dotool");
    myProcess.setProgram(program);
    myProcess.setProcessChannelMode(QProcess::ForwardedOutputChannel);
    myProcess.start(QIODeviceBase::Unbuffered | QIODeviceBase::ReadWrite);
    connect(&myProcess, SIGNAL(readyReadStandardError()), this, SLOT(doToolError()));
}

DoToolPipe::~DoToolPipe() {
    myProcess.closeWriteChannel();
    myProcess.terminate();
    myProcess.waitForFinished();
}

void DoToolPipe::doToolError() {
    QString mess = myProcess.readAllStandardError();
    std::cerr << mess.toStdString().c_str();
    if (!errorCount) {
        QString deviceError("failed to create virtual keyboard device");
        if (mess.contains(deviceError)) {
            QMessageBox::information(nullptr
                                     , QApplication::tr("%1 dotool error").arg(APP_NAME)
                                     , QApplication::tr("An error was detected running dotool.\n\nIf this is your first time using %1, you should restart your computer.\n\n%2").arg(APP_NAME).arg(mess)
                                     );
        }
        errorCount++;
    }
}

void DoToolPipe::setSteps(int moveStep_, int glideStep_, int wheelStep_) {
    moveStep  = moveStep_;
    glideStep = glideStep_;
    wheelStep = wheelStep_;
}

void DoToolPipe::writePipe(QString cmd) {
    myProcess.write(cmd.toStdString().c_str());
    myProcess.write("\n", 1);
    myProcess.waitForBytesWritten();
}

void DoToolPipe::setDelays(int keyDelay, int keyHold, int typeDelay, int typeHold) {
    writePipe(QString("keydelay %1").arg(keyDelay));
    writePipe(QString("keyhold %1").arg(keyHold));
    writePipe(QString("typedelay %1").arg(typeDelay));
    writePipe(QString("typehold %1").arg(typeHold));
}

void DoToolPipe::mouseTo(float x, float y) {
    writePipe(QString("mouseto %1 %2").arg(x).arg(y));
}

void DoToolPipe::mouseAction(Command::mouseType typ, int count) {
    for (int i = 0; i < count; i++) {
        if (typ < Command::WheelUp)
            thread()->msleep(50);      // Required for double, triple clicks
        switch (typ) {
        case Command::Left:
            writePipe("click left");
            break;
        case Command::Middle:
            writePipe("click middle");
            break;
        case Command::Right:
            writePipe("click right");
            break;
        case Command::WheelUp:
            writePipe(QString("wheel %1").arg(wheelStep));
            break;
        case Command::WheelDown:
            writePipe(QString("wheel %1").arg(-wheelStep));
            break;
        case Command::MoveLeft:
            writePipe(QString("mousemove %1 0").arg(-moveStep));
            break;
        case Command::MoveRight:
            writePipe(QString("mousemove %1 0").arg(moveStep));
            break;
        case Command::MoveUp:
            writePipe(QString("mousemove 0 %1").arg(-moveStep));
            break;
        case Command::MoveDown:
            writePipe(QString("mousemove 0 %1").arg(moveStep));
            break;

        case Command::GlideLeft:
            writePipe(QString("mousemove %1 0").arg(-glideStep));
            break;
        case Command::GlideRight:
            writePipe(QString("mousemove %1 0").arg(glideStep));
            break;
        case Command::GlideUp:
            writePipe(QString("mousemove 0 %1").arg(-glideStep));
            break;
        case Command::GlideDown:
            writePipe(QString("mousemove 0 %1").arg(glideStep));
            break;

        case Command::DownLeft:
            writePipe("buttondown left");
            break;
        case Command::DownMiddle:
            writePipe("buttondown middle");
            break;
        case Command::DownRight:
            writePipe("buttondown right");
            break;
        case Command::UpLeft:
            writePipe("buttonup left");
            break;
        case Command::UpMiddle:
            writePipe("buttonup middle");
            break;
        case Command::UpRight:
            writePipe("buttonup right");
            break;
        default:
            qWarning() << "Type not handled:" << typ;
        }
    }
}

void DoToolPipe::sendSequence(QString seq) {
    QStringList keys = seq.split(("+"));
    for (int i = 0; i < keys.count(); i++) {
        if (qtToKeysym.contains(keys[i]))
            keys[i] = qtToKeysym[keys[i]];
        else
            keys[i] = keys[i].toLower();
    }
    writePipe(QString("key %1").arg(keys.join("+")));
}
