/*
 * SPDX-FileCopyrightText: 2024-2025 Bruno Anselme <be.root@gfree.fr>
 *
 * Authors:
 *   Bruno Anselme <be.root@gfree.fr>
 *
 *   SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This file is part of the NoComprendo software.
*/

/**
 * @file      editcommand.cpp
 * @author    Bruno ANSELME
 * @brief     EditCommand class definition
 * @details   Command editor.\n
 *              Handle all commands types in a dialog.\n
 */

#include <QFileDialog>
//#include <QMetaMethod>
#include <QSettings>

#include "app_version.h"
#include "editcommand.h"

#include <QSortFilterProxyModel>
#include <QFileSystemModel>

// Custom proxy for filtering executables
// Reference: https://stackoverflow.com/questions/50213049/qfiledialog-is-it-possible-to-filter-only-executables-under-linux
class FileFilterProxyModel : public QSortFilterProxyModel {
private:
    virtual bool filterAcceptsRow(int sourceRow, const QModelIndex &sourceParent) const;
};

bool FileFilterProxyModel::filterAcceptsRow(int sourceRow, const QModelIndex &sourceParent) const {
    QFileSystemModel* fileModel = qobject_cast<QFileSystemModel*>(sourceModel());
    QFileInfo file( fileModel->filePath(sourceModel()->index(sourceRow, 0, sourceParent)) );
    if (fileModel!=NULL && file.isExecutable())
        return true;
    else
        return false;
}

//---------------
EditCommand::EditCommand(QString utt, Command &cmd, Comprendo *what_) : QDialog(),
    ui(new Ui::dialCommand)
{
    what = what_;
    lastType = Command::Meta;
    initUi();

    ui->leUtterance->setText(utt);
    switch (cmd.getType()) {
    case Command::Meta:
        ui->btMeta->click();
        ui->cbMeta->setCurrentIndex(static_cast<int>(cmd.getMetaType()));
        ui->lbMetaDesc->setText(static_cast<QString>(cmd.getMetaDescription()));
        break;
    case Command::Key:
        ui->btKey->click();
        ui->leShortcut->setText(SEQ_TO_NATIVE(cmd.getQtKey()));
        break;
    case Command::Mouse:
        ui->btMouse->click();
        ui->cbMouse->setCurrentIndex(static_cast<int>(cmd.getMouseType()));
        break;
    case Command::Text:
        ui->btText->click();
        ui->teTextInput->setText(cmd.getText());
        break;
    case Command::Exec:
        ui->btExec->click();
        ui->teCmdLine->setText(cmd.getText());
        break;
    }
    QSettings settings(APP_ORG, APP_NAME);
    move(settings.value("comPos", QPoint(200, 300)).toPoint());
    resize(settings.value("comSize", QSize(467, 283)).toSize());

    connect(ui->btRecord, SIGNAL(clicked()), this, SLOT(startRecording()));
    connect(ui->btBrowse, SIGNAL(clicked()), this, SLOT(browseExecutable()));
    connect(ui->btOk,     SIGNAL(clicked()), this, SLOT(checkAccept()));
    connect(ui->cbMeta,   SIGNAL(currentIndexChanged(int)), this, SLOT(currentMetaChanged()));
    startTimer(250);
    ui->lbUtteranceStatus->setStyleSheet("QLabel { color : red; }");
}

EditCommand::~EditCommand() {
    QSettings settings(APP_ORG, APP_NAME);
    settings.setValue("comPos",    pos());
    settings.setValue("comSize",   size());
    delete ui;
}

void EditCommand::currentMetaChanged() {
    Command *cmd = Command::createMetaCommand(static_cast<Command::metaType>(ui->cbMeta->currentIndex()));
    ui->lbMetaDesc->setText(static_cast<QString>(cmd->getMetaDescription()));
    delete cmd;
}

void EditCommand::checkAccept() {
    if (!ui->leUtterance->text().trimmed().isEmpty())
        accept();
    else
        ui->lbUtteranceStatus->setText(tr("Empty utterance"));
}

void EditCommand::browseExecutable() {
    QFileDialog dialog( this, tr("Select an executable file"), "/usr/bin");
    FileFilterProxyModel* proxyModel = new FileFilterProxyModel;
    dialog.setProxyModel(proxyModel);
    dialog.setOption(QFileDialog::DontUseNativeDialog); // required by proxy model
    QFileInfo fi(ui->teCmdLine->toPlainText().trimmed());
    dialog.selectFile(fi.fileName());
    if( dialog.exec() == QDialog::Accepted ) {
        ui->teCmdLine->setText(dialog.selectedFiles()[0]);
    }
}

void EditCommand::startRecording() {
    ui->btRecord->setEnabled(false);
    ui->stack->setEnabled(false);
    ui->gbType->setEnabled(false);
    ui->leUtterance->setEnabled(false);
    ui->leUtterance->setText(tr("Recording an utterance"));
    what->listen(true);
}

void EditCommand::timerEvent(QTimerEvent *) {
    if (what->isListening()) {
        QString *utt = what->recognize_from_microphone();
        QString utterance(utt->toStdString().c_str());
        if (utterance.length())      // Uppercase first letter
            utterance[0] = utterance[0].toUpper();

        Comprendo::VoskStatus status = what->getStatus();
        switch(status) {
        case Comprendo::VoskStatus::Partial :
            ui->leUtterance->setText(utterance);
            break;
        case Comprendo::VoskStatus::Final :
            ui->leUtterance->setEnabled(true);
            ui->leUtterance->setText(utterance);
            what->listen(false);
            ui->btRecord->setEnabled(true);
            ui->stack->setEnabled(true);
            ui->gbType->setEnabled(true);
            ui->lbUtteranceStatus->setText("");
            break;
        case Comprendo::VoskStatus::Empty :
            break;
        }
        delete utt;
    }
}

void EditCommand::initUi() {
    ui->setupUi(this);
    buts.addButton(ui->btMeta);
    buts.addButton(ui->btKey);
    buts.addButton(ui->btMouse);
    buts.addButton(ui->btText);
    buts.addButton(ui->btExec);
    buts.setId(ui->btMeta, 0);
    buts.setId(ui->btKey, 1);
    buts.setId(ui->btMouse, 2);
    buts.setId(ui->btText, 3);
    buts.setId(ui->btExec, 4);
    connect(ui->btMeta,        SIGNAL(clicked()),          this,   SLOT(toMeta()));
    connect(ui->btKey,         SIGNAL(clicked()),          this,   SLOT(toKey()));
    connect(ui->btMouse,       SIGNAL(clicked()),          this,   SLOT(toMouse()));
    connect(ui->btText,        SIGNAL(clicked()),          this,   SLOT(toText()));
    connect(ui->btExec,        SIGNAL(clicked()),          this,   SLOT(toExec()));
    connect(ui->kseqEdit,      SIGNAL(editingFinished()),  this,   SLOT(newKeySequence()));
    ui->cbMeta->addItems(Command::getMetaNames());
    ui->cbMouse->addItems(Command::getMouseNames());
}

void EditCommand::clearSignals(Command::comType newtype) {
    switch (lastType) {
    case Command::Key:
        disconnect(ui->leShortcut, SIGNAL(textChanged(const QString &)), this, SLOT(checkSequence(const QString &)));
        break;
    case Command::Exec:
        disconnect(ui->teCmdLine,  SIGNAL(textChanged()), this, SLOT(checkExec()));
        break;
    default:
        break;
    }
    lastType = newtype;
}

void EditCommand::toMeta() {
    clearSignals(Command::Meta);
    ui->gbMeta->setEnabled(true);
    ui->gbKey->setEnabled(false);
    ui->gbMouse->setEnabled(false);
    ui->gbText->setEnabled(false);
    ui->gbExec->setEnabled(false);
    ui->stack->setCurrentIndex(0);
    ui->btOk->setEnabled(true);
    currentMetaChanged();
}

void EditCommand::toKey() {
    clearSignals(Command::Key);
    ui->gbMeta->setEnabled(false);
    ui->gbKey->setEnabled(true);
    ui->gbMouse->setEnabled(false);
    ui->gbText->setEnabled(false);
    ui->gbExec->setEnabled(false);
    ui->stack->setCurrentIndex(1);
    connect(ui->leShortcut,    SIGNAL(textChanged(const QString &)), this, SLOT(checkSequence(const QString &)));
    checkSequence(ui->leShortcut->text());
}

void EditCommand::toMouse() {
    clearSignals(Command::Mouse);
    ui->gbMeta->setEnabled(false);
    ui->gbKey->setEnabled(false);
    ui->gbMouse->setEnabled(true);
    ui->gbText->setEnabled(false);
    ui->gbExec->setEnabled(false);
    ui->stack->setCurrentIndex(2);
    ui->btOk->setEnabled(true);
}

void EditCommand::toText() {
    clearSignals(Command::Text);
    ui->gbMeta->setEnabled(false);
    ui->gbKey->setEnabled(false);
    ui->gbMouse->setEnabled(false);
    ui->gbText->setEnabled(true);
    ui->gbExec->setEnabled(false);
    ui->stack->setCurrentIndex(3);
    ui->btOk->setEnabled(true);
}

void EditCommand::toExec() {
    clearSignals(Command::Exec);
    ui->gbMeta->setEnabled(false);
    ui->gbKey->setEnabled(false);
    ui->gbMouse->setEnabled(false);
    ui->gbText->setEnabled(false);
    ui->gbExec->setEnabled(true);
    ui->stack->setCurrentIndex(4);
    connect(ui->teCmdLine,    SIGNAL(textChanged()), this, SLOT(checkExec()));
    checkExec();
}

Command::comType    EditCommand::getType()  { return static_cast<Command::comType>(buts.checkedId());               }
Command::metaType   EditCommand::getMeta()  { return static_cast<Command::metaType>(ui->cbMeta->currentIndex());    }
Command::mouseType  EditCommand::getMouse() { return static_cast<Command::mouseType>(ui->cbMouse->currentIndex());  }

QString         EditCommand::getText()      { return ui->teTextInput->toPlainText();   }
QString         EditCommand::getExec()      { return ui->teCmdLine->toPlainText().trimmed();     }
QKeySequence    EditCommand::getSequence()  { QKeySequence  seq(ui->leShortcut->text()); return seq;    }

void EditCommand::newKeySequence() {
    ui->leShortcut->setText(ui->kseqEdit->keySequence().toString(QKeySequence::NativeText));
    checkSequence(ui->leShortcut->text());
}

void EditCommand::checkSequence(const QString &s) {
    QKeySequence  seq(s);
    if ((!s.isEmpty()) && (s == seq.toString(QKeySequence::NativeText))) {
        ui->lbKeyStatus->setStyleSheet("QLabel {}");
        ui->lbKeyStatus->setText(tr("Valid shortcut"));
        ui->btOk->setEnabled(true);
    } else {
        ui->lbKeyStatus->setStyleSheet("QLabel { color : red; }");
        ui->lbKeyStatus->setText(tr("Invalid shortcut"));
        ui->btOk->setEnabled(false);
    }
}

void EditCommand::checkExec() {
    QStringList parts =  ui->teCmdLine->toPlainText().trimmed().split(QRegularExpression("\\s+"));
    QFileInfo info(parts.first());
    if ((!info.isDir()) && info.isExecutable()) {
        ui->lbExecStatus->setStyleSheet("QLabel {}");
        ui->lbExecStatus->setText(tr("<b>%1</b> is an executable file").arg(parts.first()));
        ui->btOk->setEnabled(true);
    } else {
        ui->lbExecStatus->setStyleSheet("QLabel { color : red; }");
        ui->lbExecStatus->setText(tr("<b>%1</b> is not an executable file").arg(parts.first()));
        ui->btOk->setEnabled(false);
    }
}

