﻿/*
 * SPDX-FileCopyrightText: 2024-2025 Bruno Anselme <be.root@gfree.fr>
 *
 * Authors:
 *   Bruno Anselme <be.root@gfree.fr>
 *
 *   SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This file is part of the NoComprendo software.
*/

#include <iostream>
#include <signal.h>
#include <sys/socket.h>
#include <unistd.h>

#include <QInputDialog>
#include <QMessageBox>
#include <QFileDialog>
#include <QDebug>
#include <QClipboard>
#include <QMimeData>
#include <QActionGroup>

#include "app_version.h"
#include "ui_mainwindow.h"
#include "mainwindow.h"
#include "settings.h"
#include "settab.h"
#include "misc.h"
#include "modelloader.h"

static int  sigtermFd[2];
static int  sigintFd[2];
static int  sighupFd[2];

static int setup_unix_signal_handlers()
{
    struct sigaction sigterm;
    sigterm.sa_handler = MainWindow::termSignalHandler;
    sigemptyset(&sigterm.sa_mask);
    sigterm.sa_flags |= SA_RESTART;
    if (sigaction(SIGTERM, &sigterm, 0))
       return SIGTERM;

    struct sigaction sigint;
    sigint.sa_handler = MainWindow::intSignalHandler;
    sigemptyset(&sigint.sa_mask);
    sigint.sa_flags |= SA_RESTART;
    if (sigaction(SIGINT, &sigint, 0))
       return SIGINT;

    struct sigaction sighup;
    sighup.sa_handler = MainWindow::hupSignalHandler;
    sigemptyset(&sighup.sa_mask);
    sighup.sa_flags |= SA_RESTART;
    if (sigaction(SIGHUP, &sighup, 0))
       return SIGHUP;

    return 0;
}

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent)
    , ui(new Ui::MainWindow)
    , what(nullptr)
    , firstPartial(true)
    , listening(false)
    , dictation(false)
    , dictLength(0)
    , lastDict("")
    , once(false)
    , lastPrintedChar(0)
    , shortcutFormat(CommandSet::QtShortcut)
    , reminder(nullptr)
    , listenState(nullptr)
    , timerid(0)
    , suspended(false)
    , lastOk(true)
    , lineCount(0)
    , tray(nullptr)
    , osd(nullptr)
    , blinker(nullptr)
    , backSpace(" backspace")
    , speech()
    , speaking(false)
    , themes( { "color-light", "color-dark", "light", "dark" } )
{
    // Connect TERM signal handlers
    if (::socketpair(AF_UNIX, SOCK_STREAM, 0, sigtermFd))
       qFatal("Couldn't create TERM socketpair");
    snTerm = new QSocketNotifier(sigtermFd[1], QSocketNotifier::Read, this);
    connect(snTerm, SIGNAL(activated(int)), this, SLOT(handleSigTerm()));
    // Connect INT signal handlers
    if (::socketpair(AF_UNIX, SOCK_STREAM, 0, sigintFd))
       qFatal("Couldn't create INT socketpair");
    snInt = new QSocketNotifier(sigintFd[1], QSocketNotifier::Read, this);
    connect(snInt, SIGNAL(activated(int)), this, SLOT(handleSigInt()));
    // Connect HUP signal handlers
    if (::socketpair(AF_UNIX, SOCK_STREAM, 0, sighupFd))
       qFatal("Couldn't create HUP socketpair");
    snHup = new QSocketNotifier(sighupFd[1], QSocketNotifier::Read, this);
    connect(snHup, SIGNAL(activated(int)), this, SLOT(handleSigHup()));

    setup_unix_signal_handlers();

    setWindowTitle(APP_NAME);
    aboutDialog = nullptr;
    Command::initCommands();
    ui->setupUi(this);
    // First time actions : if directory is empty, copy default commands sets, create settings file
    if (APPCONF.userSetsPath.isEmpty())
        copyExamples(false);

    what        = new Comprendo(this);
    what->setRepeaters(readFile(APPCONF.repeaterFile).split("\n", Qt::SkipEmptyParts));

    updateSets();
    readSettings();
    what->restartDotool();                  // Now we've got a keyboardVariant
    what->setAudioDevice(APPCONF.audioId);  // Now we've got an audio source

    delayedMethod   = nullptr;
    runningMethod   = nullptr;

    // Setup status bar
    QLabel *sourcesLabel = new QLabel(tr("Audio input"));
    statusBar()->addPermanentWidget(sourcesLabel);

    cbSources = new QComboBox();
    const QList<QAudioDevice> audioDevices = QMediaDevices::audioInputs();
    int idx = 0;
    for (const QAudioDevice &device : audioDevices) {
        cbSources->addItem(device.description(), device.id());
        if (APPCONF.audioId == device.id())
            cbSources->setCurrentIndex(idx);
        idx++;
    }
    cbSources->setMaximumWidth(200);
    statusBar()->addPermanentWidget(cbSources);

    listenState     = new QLabel(tr("Stopped"));
    QFont font1;
    font1.setBold(true);
    listenState->setFont(font1);
    listenState->setMinimumWidth(100);
    listenState->setAlignment(Qt::AlignRight | Qt::AlignVCenter);
    statusBar()->addPermanentWidget(listenState);

    // Setup OSD and blinker
    osd = new OSDWindow();
    blinker = new MouseBlink();

    // Read utterances for multiplier
    multiplierUtterance = { "once", "twice", "three times", "four times", "five times", "six times", "seven times", "eight times", "nine times", "ten times"};
    QString txt =   readFile(APPCONF.repeaterFile);
    if ((!txt.isEmpty()) && (txt.count("\n") == 10))
        multiplierUtterance = txt.split("\n");

    show();
    if (APPCONF.startReduced) {
        setWindowState((windowState() | Qt::WindowMinimized));
    } else {
        activateWindow();
    }

    connect(ui->btQuit,      SIGNAL(clicked()),      this,      SLOT(askForClose()));
    connect(ui->btPlay,      SIGNAL(toggled(bool)),  this,      SLOT(toggleListening(bool)));
    connect(ui->btSuspend,   SIGNAL(toggled(bool)),  this,      SLOT(toggleSuspend(bool)));
    connect(ui->btDictation, SIGNAL(clicked(bool)),  this,      SLOT(toggleDictation(bool)));
    connect(ui->btClear,     SIGNAL(clicked()),      this,      SLOT(clearConsole()));
    connect(ui->btConfig,    SIGNAL(clicked()),      this,      SLOT(doSettings()));
    connect(ui->btReminder,  SIGNAL(clicked()),      this,      SLOT(openReminder()));
    connect(what,            SIGNAL(statusListening(bool)),     this, SLOT(listeningStatus(bool)));
    connect(ui->btNewSet,    SIGNAL(clicked()),      this,      SLOT(newSet()));
    connect(ui->btRenameSet, SIGNAL(clicked()),      this,      SLOT(renameSet()));
    connect(ui->btDeleteSet, SIGNAL(clicked()),      this,      SLOT(deleteSet()));
    connect(ui->btImportSet, SIGNAL(clicked()),      this,      SLOT(importSet()));
    connect(ui->lstSets,     SIGNAL(itemChanged(QListWidgetItem*)), this,   SLOT(updateCurrentSet(QListWidgetItem*)));
    connect(ui->btAbout,     SIGNAL(clicked()),      this,      SLOT(about()));
    connect(ui->btSpeech,    SIGNAL(clicked()),      this,      SLOT(speak()));
    connect(cbSources,       SIGNAL(currentIndexChanged(int)), this, SLOT(changeAudioSource(int)));

    runSysTray();
    if (APPCONF.autoStart)
        ui->btPlay->click();
    if (tray)
        tray->setMenus((timerid != 0), suspended);
    ui->btDictation->setEnabled(APPCONF.autoStart);
    ui->btSpeech->setToolTip(tr("Read copied text"));
    ui->btSpeech->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(ui->btSpeech, SIGNAL(customContextMenuRequested(QPoint)), this, SLOT(showContextMenu(QPoint)));

//    APPCONF.trace();    // Trace when everything is ready (after ReadSetttings)
}

MainWindow::~MainWindow() {
    writeSettings();
    delete blinker;
    delete osd;
    if (tray) delete tray;
    delete what;
    delete ui;
}

void MainWindow::changeAudioSource(int) {
    what->setAudioDevice(cbSources->currentData().toByteArray());
}

void MainWindow::showContextMenu(QPoint where) {
    if (speech.isRunning())
        return;
    QPoint globalPos = ui->btSpeech->mapToGlobal(where);
    /* Create menu and insert some actions */
    QMenu speechListMenu(this);
    QActionGroup actionGroup(this);
    actionGroup.setExclusive(true);

    QMapIterator<QString, QString> m(APPCONF.speechesList);
    while (m.hasNext()) {
        m.next();
        QAction* action= speechListMenu.addAction(m.value());
        action->setData(m.key());
        action->setCheckable(true);
        if (m.key() == APPCONF.speechLang)
            action->setChecked(true);
        actionGroup.addAction(action);
    }

    QAction *selectedMenuItem = speechListMenu.exec(globalPos);
    if (selectedMenuItem)
        APPCONF.speechLang = selectedMenuItem->data().toString();
}

// Run system tray if available. Connect menus signals.
void MainWindow::runSysTray() {
    if ((APPCONF.useSysTray) && (QSystemTrayIcon::isSystemTrayAvailable())) {
        if (!tray) {
            tray = new TrayIcon();
            connect(tray, SIGNAL(trayDictation()),  ui->btDictation,    SLOT(click()));
            connect(tray, SIGNAL(traySpeak()),      ui->btSpeech,       SLOT(click()));
            connect(tray, SIGNAL(trayStart()),      ui->btPlay,         SLOT(click()));
            connect(tray, SIGNAL(trayStop()),       ui->btPlay,         SLOT(click()));
            connect(tray, SIGNAL(traySuspend()),    ui->btSuspend,      SLOT(click()));
            connect(tray, SIGNAL(trayQuit()),       ui->btQuit,         SLOT(click()));
            connect(tray, SIGNAL(trayShow()),       this,               SLOT(raiseWindow()));
            connect(tray, SIGNAL(activated(QSystemTrayIcon::ActivationReason)), this, SLOT(raiseWindow()));
        }
    } else {
        if (tray) {
            delete tray;        // disconnect all signals
            tray = nullptr;
        }
    }
}

void MainWindow::about() {
    if (aboutDialog == nullptr) {
        aboutDialog = new AboutDialog(APPCONF.lang, this);
    }
    aboutDialog->show();
    aboutDialog->raise();
    aboutDialog->activateWindow();
}

void  MainWindow::updateCurrentSet(QListWidgetItem *item) {
    ui->lstSets->setCurrentItem(item);
    ui->setTabs->setTabVisible(ui->lstSets->currentRow() + 1, item->checkState() == Qt::Checked);
    QSettings settings(APP_ORG, APP_NAME);
    settings.remove(APPCONF.setsDirName);
    settings.beginGroup(APPCONF.setsDirName);
    for (int i = 0; i < ui->lstSets->count(); i++) {
        QListWidgetItem *item = ui->lstSets->item(i);
        if (item->checkState() & Qt::Checked) {
            settings.setValue(item->text(), item->text().append(".noc"));
        }
    }
    settings.endGroup();
    what->loadAllSets();
}

void MainWindow::updateSets() {
    ui->lstSets->clear();
    while (ui->setTabs->count() > 1)    // Clear sets tabs.
        ui->setTabs->removeTab(1);
    QStringList filter = { "*.noc" };
    QStringList files = APPCONF.userSetsPath.entryList(filter);
    QStringList keys = what->getSetFiles(APPCONF.setsDirName);
    QListIterator<QString> fileIterator(files);
    QStringList setNames;
    int counter = 1;
    while (fileIterator.hasNext()) {
        QString name = fileIterator.next();
        name.replace(QRegularExpression("\\.noc$"),"");
        setNames << name;
        QListWidgetItem *item = new QListWidgetItem(name);
        if (keys.contains(name))
            item->setCheckState(Qt::Checked);
        else
            item->setCheckState(Qt::Unchecked);
        ui->lstSets->addItem(item);

        // Add settab here
        SetTab *newTab = new SetTab(what, name, ui->setTabs);
        ui->setTabs->addTab(newTab, name);
        ui->setTabs->setTabVisible(counter, item->checkState() == Qt::Checked);
        connect(newTab, SIGNAL(setListening(bool)), this, SLOT(forceListening(bool)));
        counter++;
    }
    what->loadAllSets();
}

void MainWindow::newSet() {
    bool ok;
    QString setName = QInputDialog::getText(this, tr("New commands set"),
                                         tr("New commands set name :"), QLineEdit::Normal,
                                         tr("untitled"), &ok);
    if (ok && !setName.isEmpty()) {
        QString dstName    = QString("%1/%2.noc").arg(APPCONF.userSetsPath.absolutePath()).arg(setName);
        QFileInfo check_file(dstName);
        if (check_file.exists() && check_file.isFile()) {
            QMessageBox msgBox;
            msgBox.setText(tr("Commands set <b>%1</b> already exists.").arg(setName));
            msgBox.exec();
        } else {
            QFile file(dstName);
            if (!file.open(QIODevice::WriteOnly | QIODevice::Text))
                return;
            file.close();
            SetTab *newTab = new SetTab(what, setName, ui->setTabs);
            ui->setTabs->addTab(newTab, setName);

            QSettings settings(APP_ORG, APP_NAME);
            settings.beginGroup(APPCONF.setsDirName);
            QString fileName = setName;
            fileName.append(".noc");
            settings.setValue(setName, fileName);
            settings.endGroup();

            connect(newTab, SIGNAL(setListening(bool)), this, SLOT(forceListening(bool)));
            updateSets();
        }
    }
}

void MainWindow::renameSet() {
    if (ui->lstSets->currentRow() == -1)
        return;
    bool ok;
    QString text = QInputDialog::getText(this, tr("Rename set"),
                                         tr("New set name for <b>%1</b>:").arg(ui->lstSets->currentItem()->text()), QLineEdit::Normal,
                                         ui->lstSets->currentItem()->text(), &ok);
    if (ok && !text.isEmpty()) {
        QString oldName = QString("%1/%2.noc").arg(APPCONF.userSetsPath.path()).arg(ui->lstSets->currentItem()->text());
        QString newName = QString("%1/%2.noc").arg(APPCONF.userSetsPath.path()).arg(text);
        if (QFile::exists(newName)) {
            QString mess = tr("Set <b>%1</b> already exists").arg(text);
            information(mess);
        } else {
            QFile::rename(oldName, newName);
            updateSets();
        }
    }
}

void MainWindow::deleteSet() {
    if (ui->lstSets->currentRow() == -1)
        return;
    QMessageBox::StandardButton reply;
    reply = QMessageBox::question(this,
                                  tr("Delete set"),
                                  tr("Delete set : <b>%1</b>").arg(ui->lstSets->currentItem()->text()),
                                  QMessageBox::Yes|QMessageBox::No, QMessageBox::No);
    if (reply == QMessageBox::Yes) {
        QString srcName    = QString("%1/%2.noc").arg(APPCONF.userSetsPath.path()).arg(ui->lstSets->currentItem()->text());
        QFile::remove(srcName);
        updateSets();
    }
}

void MainWindow::raiseWindow() {        // This function comes into conflict with desktop manager
    setWindowState((windowState() & ~Qt::WindowMinimized));
    show();
    activateWindow();
    raise();
}

void MainWindow::askForClose() {
    if (APPCONF.ask2Quit) {
        QMessageBox::StandardButton reply = QMessageBox::question(nullptr, tr("Exit"), tr("Do you want to quit %1 ?").arg(APP_NAME),
                QMessageBox::Yes|QMessageBox::No, QMessageBox::No);
        if (reply == QMessageBox::No)
            return;
    }
    close();
}

void MainWindow::forceListening(bool b) {
    if (b != ui->btPlay->isChecked())
        ui->btPlay->setChecked(b);
}
void MainWindow::toggleDictation(bool b) {
    dictation = b;
    ui->btDictation->setChecked(dictation);
    if (tray) {
        tray->setDictationChecked(dictation);
        tray->setStatus(dictation ? TrayIcon::Dictation : TrayIcon::Ready);
    }
    listenState->setText(dictation ? tr("Dictation") : tr("Listening"));
}

void MainWindow::toggleListening(bool b) {
    listening = b;
    ui->btDictation->setEnabled(b);
    what->listen(b);    // Start or stop recognizer
    if (b) {
        firstPartial = true;
        ui->textOutput->setFontWeight(QFont::Bold);
//        ui->textOutput->append(tr("Ready..."));
        ui->textOutput->setFontWeight(QFont::Normal);
        lineCount++;
        if (APPCONF.osdOn) osd->setFlashingText(tr("Ready..."));
        timerid = startTimer(250);
        listenState->setText(tr("Listening"));
        ui->btSuspend->setEnabled(true);
        ui->btSuspend->setChecked(false);
        if (tray) tray->setStatus(TrayIcon::Ready);
    } else {
        killTimer(timerid);
        suspended = false;
        timerid = 0;
        listenState->setText(tr("Stopped"));
        statusBar()->showMessage(tr("Stopped"));
        ui->btSuspend->setChecked(false);
        ui->btSuspend->setEnabled(false);
        ui->textOutput->setFontWeight(QFont::Bold);
//        ui->textOutput->append(tr("Stopped"));
        ui->textOutput->setFontWeight(QFont::Normal);
        if (APPCONF.osdOn) osd->setFlashingText(tr("Stopped"));
        lineCount++;
        if (tray) tray->setStatus(TrayIcon::Stopped);
    }
    if (tray) tray->setMenus((timerid != 0), suspended);
    lastOk = true;
}

void MainWindow::toggleSuspend(bool b) {
    ui->textOutput->setFontWeight(QFont::Bold);
    if (b) {
        suspended = true;
        listenState->setText(tr("Suspended"));
        ui->btDictation->setEnabled(false);
        if (APPCONF.osdOn) osd->setFlashingText(tr("Suspended"));
        if (tray)  tray->setStatus(TrayIcon::Suspended);
    } else {
        suspended = false;
        listenState->setText(dictation ? tr("Dictation") : tr("Listening"));
        ui->btDictation->setEnabled(true);
        if (APPCONF.osdOn) osd->setFlashingText(tr("Listening"));
        if (tray)  tray->setStatus(TrayIcon::Ready);
    }
    ui->textOutput->setFontWeight(QFont::Normal);
    lastOk = true;
    if (tray) tray->setMenus((timerid != 0), suspended);
}

void MainWindow::timerEvent(QTimerEvent *t) {
    int pos = 0;
    QString copy;
    if (!listening)
        return;
    if (what->isListening()) {
        if (t->timerId()==timerid) {
            QString *utt = what->recognize_from_microphone();
            if (utt->length() == 0) {
                delete utt;
                return;
            }
            if (!dictation || APPCONF.capitalizeFirst)
                (*utt)[0] = (*utt)[0].toUpper();        // Uppercase first letter
            Comprendo::VoskStatus status = what->getStatus();
            switch(status) {
            case Comprendo::VoskStatus::Partial :
                if (firstPartial) {
                    firstPartial = false;
                } else {
                    if (!suspended && dictation && APPCONF.transparentMode) {
                        pos = lastCommonChar(&lastDict, utt);
                        if (dictLength - pos > 0)
                            what->writeRaw(QString("key%1").arg(backSpace.repeated(dictLength - pos)));
                        lastDict = *utt;
                    }
                    ui->textOutput->undo();
                }
                ui->textOutput->setTextColor(Qt::darkGray);
                ui->textOutput->append(*utt);
                if (!suspended && dictation && APPCONF.transparentMode) {
                    copy = *utt;
                    if (lastPrintedChar == QChar(','))
                        copy[0] = copy[0].toLower();
                    what->writeText(copy.sliced(pos));
                }
                dictLength = utt->length();
                break;
            case Comprendo::VoskStatus::Final :
                if (!firstPartial)
                    ui->textOutput->undo();
                firstPartial = true;
                if (!suspended && dictation && APPCONF.transparentMode) {
                    pos = lastCommonChar(&lastDict, utt);
                    if (dictLength - pos > 0)
                        what->writeRaw(QString("key%1").arg(backSpace.repeated(dictLength - pos)));
                    copy = *utt;
                    if (lastPrintedChar == QChar(','))
                        copy[0] = copy[0].toLower();
                    what->writeText(copy.sliced(pos));
                }
                dictLength = utt->length();
                utteranceReceived(*utt);
                break;
            case Comprendo::VoskStatus::Empty :
                break;
            }
            delete utt;        // Utterance pointer created in recognize_from_microphone must be deleted here
        }
    }
}

int MainWindow::lastCommonChar(QString *s1, QString *s2) {
    int pos = 0;
    int max = qMin(s1->length(), s2->length());
    for (int i = 0; i < max; i++) {
        if (s1->at(i) != s2->at(i))
            break;
        pos++;
    }
    return  pos;
}

void MainWindow::utteranceReceived(QString utterance) {
    bool wassuspended   = suspended;
    bool italic = dictation;
    int count = 1;

    QString line;       // Displayed by ui->textOutput and osd
    Command    *com;
    com = what->getCommand(utterance, count);
    bool cmdok = (com) ? true : false;      // Last utterance was an existing command
    if (com) {
        italic = false;
        if (!suspended && dictation && APPCONF.transparentMode) {
            if (dictLength)
                what->writeRaw(QString("key%1").arg(backSpace.repeated(dictLength)));
        }
        if (com->getType() == Command::Meta) {
            if (com->getMetaType() != Command::Repeat) count = 1;  // Other meta commands don't use multipliers
            for (int j = 0; j < count; j++) {
                execMeta(com->getMetaType());
            }
        }
        if (com->getType() == Command::Text) {
            QString copy = com->getText();
            lastPrintedChar = (copy.size() > 1) ? copy.at(copy.size() - 2) : QChar(' ');    // Remember for ','
        }
        line = what->execCommand(utterance, com, count, suspended);
    } else {
        if (!APPCONF.capitalizeFirst)
            utterance[0] = utterance[0].toLower();          // Lowercase first letter
        line = utterance;
        if (!suspended && dictation) {
            if (!APPCONF.transparentMode) {
                if (lastPrintedChar == QChar(','))          // Lowercase after ','
                    utterance[0] = utterance[0].toLower();
                what->writeText(utterance);
            }
            if (once) {
                dictation = false;
                once = false;
            }
            what->writeLongText(APPCONF.dictationSuffix);
            lastPrintedChar = QChar();
        }
    }
    dictLength = 0;

    if ((!lastOk) && (APPCONF.removeBadUtterance))
        removeLastOutput();
    lastOk = cmdok || italic;
    if (!utterance.isEmpty() && ((!suspended) || (!wassuspended))) {     // Stop displaying while suspended
        ui->textOutput->setFontWeight(QFont::Normal);
        if (cmdok && APPCONF.commandTrace) {
            std::cout << utterance.toStdString().c_str() << "\n";
            std::cout.flush();
        }
        if (!cmdok) ui->textOutput->setTextColor(Qt::lightGray);    // unknown utterance color
        else        ui->textOutput->setTextColor(Qt::black);        // valid utterance colorOui
        if (italic)
            ui->textOutput->setFontItalic(true);
        ui->textOutput->append(line);
        if (italic)
            ui->textOutput->setFontItalic(false);
        ui->textOutput->moveCursor(QTextCursor::End);
        ui->textOutput->moveCursor(QTextCursor::StartOfBlock);
        ui->textOutput->setTextColor(Qt::black);                    // back to default color
        lineCount++;
        ui->leRepeat->setText(what->getLastCommand());
        if (italic) {
            if (!APPCONF.osdNoDictation && APPCONF.osdOn)
                osd->setItalic(line);
        } else if ((cmdok) && (APPCONF.osdOn))
            osd->setFlashingText(line);
        else if ((APPCONF.osdOn) && (APPCONF.osdAll))
            osd->setUnknown(utterance);
        if (delayedMethod) {                    // These methods must start from top level loop.
            runningMethod = delayedMethod;      // Copy method to an other pointer
            delayedMethod = nullptr;            // Clear delayedMethod pointer first to avoid loop
            (this->*runningMethod)();           // Run the delayed method (dialog) from the other pointer
            runningMethod    = nullptr;         // Clear runningMethod after execution. Ready for next delayedMethod
        }
        while (lineCount > APPCONF.maxlines) {   // Remove first lines
            QTextCursor cursor = ui->textOutput->textCursor();
            cursor.movePosition(QTextCursor::Start);
            cursor.movePosition(QTextCursor::Down, QTextCursor::KeepAnchor, 1);
            cursor.removeSelectedText();
            ui->textOutput->moveCursor(QTextCursor::End);
            lineCount--;
        }
    }
}

// Change SystemTrayIcon depending on listening and suspend states
void MainWindow::listeningStatus(bool b) {
    if (b) {
        statusBar()->showMessage(tr("Listening..."));
        if (tray) {
            if (suspended)      tray->setStatus(TrayIcon::ListenSuspend);
            else if (dictation) tray->setStatus(TrayIcon::ListenDictation);
            else                tray->setStatus(TrayIcon::Listening);
        }
    } else {
        statusBar()->showMessage(tr("Ready..."));
        if (tray) {
            if (suspended)      tray->setStatus(TrayIcon::Suspended);
            else if (dictation) tray->setStatus(TrayIcon::Dictation);
            else                tray->setStatus(TrayIcon::Ready);
        }
    }
}

void MainWindow::clearConsole() {
    ui->textOutput->clear();
    lineCount   = 0;
}

void MainWindow::removeLastOutput() {
    if (!suspended) {
        QTextCursor cursor = ui->textOutput->textCursor();
        cursor.movePosition(QTextCursor::End);
        cursor.select(QTextCursor::LineUnderCursor);
        cursor.removeSelectedText();
        cursor.deletePreviousChar();                                    // Added to trim the newline char when removing last line
        ui->textOutput->setTextCursor(cursor);
        lineCount--;
    }
}

void MainWindow::execMeta(Command::metaType m) {            // Meta-commands are executed here. Out of Comprendo object.
    switch (m) {
    case Command::Quit :
        if (!runningMethod)
            delayedMethod = &MainWindow::askForClose;
        break;
    case Command::Stop :
        if (listening)
            ui->btPlay->click();
        if (ui->btSuspend->isChecked())
            ui->btSuspend->click();
        break;
    case Command::Suspend :
        if (!ui->btSuspend->isChecked())
            ui->btSuspend->click();
        break;
    case Command::Resume :
        if (ui->btSuspend->isChecked())
            ui->btSuspend->click();
        break;
    case Command::Repeat :
        what->repeatLastCommand(suspended);
        break;
    case Command::Settings:
        if (!runningMethod)
            delayedMethod = &MainWindow::doSettings;
        break;
    case Command::Reminder:
        openReminder();
        break;
    case Command::Raise:        // Something here comes into conflict with desktop manager
        setWindowState( (windowState() & ~Qt::WindowMinimized));
        show();
        activateWindow();
        raise();
        break;
    case Command::Minimize:
        if (windowState() & Qt::WindowMinimized)
            setWindowState((windowState() & ~Qt::WindowMinimized));
        else
            setWindowState((windowState() | Qt::WindowMinimized));
        break;
    case Command::Maximize:
        if (windowState() & Qt::WindowMaximized)
            setWindowState((windowState() & ~Qt::WindowMaximized));
        else
            setWindowState((windowState() | Qt::WindowMaximized));
        break;
    case Command::MouseShow:
        blinker->start();
        break;
    case Command::StartDictation:
        toggleDictation(true);
        break;
    case Command::StopDictation:
        toggleDictation(false);
        break;
    case Command::OneDictation:
        if (!dictation) {
            dictation = true;
            once = true;
        }
        break;
    case Command::SetsSelection:
        if (!runningMethod)
            delayedMethod = &MainWindow::raiseSetSelector;
        break;
    case Command::Speak:
        speak();
        break;
    }
}

void MainWindow::speak() {
    if (!speaking) {
        what->listen(false);
        speaking = true;
        ui->btSpeech->setIcon(QIcon(":/icons/speaker-stop.png"));
        ui->btSpeech->setToolTip(tr("Stop reading"));
        connect(&speech, SIGNAL(speechFinished()), this, SLOT(resumeListening()));
        speech.setSpeech(QApplication::clipboard()->text(QClipboard::Clipboard));
        speech.start();
    } else {
        speech.stopSpeech();
        resumeListening();
    }
}

void MainWindow::resumeListening() {
    speaking = false;
    ui->btSpeech->setIcon(QIcon(":/icons/speaker.png"));
    ui->btSpeech->setToolTip(tr("Read copied text"));
    disconnect(&speech, SIGNAL(speechFinished()), this, SLOT(resumeListening()));
    what->listen(true);
}

void MainWindow::raiseSetSelector() {
    execMeta(Command::Raise);
    ui->setTabs->setCurrentIndex(0);
    ui->lstSets->setFocus();
}

void MainWindow::openReminder() {
    if (!reminder) {
        reminder = new Reminder(what);
        connect(reminder, SIGNAL(rejected()), this, SLOT(closeReminder()));
        connect(reminder, SIGNAL(accepted()), this, SLOT(closeReminder()));
        reminder->show();
    }
}

void MainWindow::closeReminder() {
    disconnect(reminder, SIGNAL(rejected()), this, SLOT(closeReminder()));
    disconnect(reminder, SIGNAL(accepted()), this, SLOT(closeReminder()));
    reminder->deleteLater();
    reminder    = nullptr;
}

void MainWindow::information(const QString &mes) {
    QMessageBox msgBox;
    msgBox.setWindowTitle(tr("%1 information").arg(APP_NAME));
    msgBox.setText(mes);
    msgBox.exec();
}

void MainWindow::doSettings() {
    bool memListening = ui->btPlay->isChecked();
    forceListening(false);
    Settings dial(what);
    connect(&dial,   SIGNAL(sendInformation(const QString &)),  this,   SLOT(information(const QString &)));
    connect(&dial,   SIGNAL(resetDefaultSets()),                this,   SLOT(copyExamples()));
    int dialogCode = dial.exec();
    if (dialogCode == QDialog::Accepted) {
        writeSettings();
        what->setRepeaters(readFile(APPCONF.repeaterFile).split("\n", Qt::SkipEmptyParts));
        what->setMouseSteps(APPCONF.mouseStep, APPCONF.glideStep, APPCONF.wheelStep);
        what->setKeyDelays(APPCONF.keyDelay, APPCONF.keyHold, APPCONF.typeDelay, APPCONF.typeHold);
        runSysTray();
        osd->changeOsd(APPCONF.osdPosition, APPCONF.osdSize);
        // Set theme here
        if (tray)
            tray->setTheme(themes[APPCONF.themeIndex]);
    }
    forceListening(memListening);
}

void MainWindow::importSet() {
    QString fileName = QFileDialog::getOpenFileName(this, tr("Open File"),
                                                    QString("%1/").arg(QDir::homePath()),
                                                    QString("%1 (*.noc)").arg(APP_NAME));
    if (!fileName.isEmpty()) {
        QFileInfo info(fileName);
        QString txt = readFile(fileName);
        QString dstname    = QString("%1/%2.noc").arg(APPCONF.userSetsPath.absolutePath()).arg(info.baseName());
        QFileInfo check_file(dstname);
        if (check_file.exists() && check_file.isFile()) {
            QMessageBox msgBox;
            msgBox.setText(tr("Commands set <b>%1</b> already exists.\nRename file.").arg(info.baseName()));
            msgBox.exec();
        } else {
            writeFile(dstname,txt);                     // New commands set
            updateSets();                               // Update this dialog
        }
    }
}

// copy default commands sets, create settings file
void MainWindow::copyExamples(bool confirm) {
    QDir    source  = QString("%1/%2").arg(APPCONF.appPath.absolutePath()).arg(APPCONF.setsDirName);
    source.setNameFilters(QStringList() << "*.noc");
    source.setFilter(QDir::Files);
    QStringList files   = source.entryList();

    if (confirm) {
        files.replaceInStrings(QRegularExpression("\\.noc"), "");
        QMessageBox::StandardButton reply;
        reply = QMessageBox::question(this,
                                      tr("Confirm"),
                                      tr("Restore default sets ?\n%1\nRepeaters and MouseMoveTo").arg(files.join(", ")),
                                      QMessageBox::Yes|QMessageBox::No, QMessageBox::No);
        if (reply == QMessageBox::No)
            return;
    }
    QSettings settings(APP_ORG, APP_NAME);
    settings.remove(APPCONF.setsDirName);
    settings.beginGroup(APPCONF.setsDirName);

    source.setNameFilters(QStringList() << "*.noc*");
    files   = source.entryList();
    QRegularExpression  cmdpat("\\.noc$");
    QListIterator<QString> filename(files);
    while (filename.hasNext()) {                // Copy all files
        QString name = filename.next();
        QString oldName = QString("%1/%2").arg(source.absolutePath()).arg(name);
        QString newName = QString("%1/%2").arg(APPCONF.userSetsPath.absolutePath()).arg(name);
        if (QFile::exists(newName))
            QFile::remove(newName);             // Remove before copy
        QFile::copy(oldName, newName);
        if (name.contains(cmdpat))              // Add commands sets to settings
            settings.setValue(name.left(name.length()-4), name);
    }
    settings.endGroup();
    if (what)
        updateSets();
}

void MainWindow::readSettings() {
    QSettings settings(APP_ORG, APP_NAME);
    move(settings.value("appPos", QPoint(200, 200)).toPoint());
    resize(settings.value("appSize", QSize(600, 450)).toSize());
    ui->splitter->restoreState(settings.value("appSplit").toByteArray());
    APPCONF.autoStart           = settings.value("autoStart", false).toBool();
    APPCONF.ask2Quit            = settings.value("ask2Quit", false).toBool();
    APPCONF.maxlines            = settings.value("maxlines", 100).toInt();
    APPCONF.startReduced        = settings.value("startReduced", false).toBool();
    APPCONF.removeBadUtterance  = settings.value("removeBadUtterance", false).toBool();
    APPCONF.transparentMode     = settings.value("transparentMode", false).toBool();
    APPCONF.osdNoDictation      = settings.value("osdNoDictation", true).toBool();
    APPCONF.capitalizeFirst     = settings.value("capitalizeFirst", false).toBool();
    APPCONF.dictationSuffix     = settings.value("dictationSuffix", "").toString();
    APPCONF.useSysTray          = settings.value("useSysTray", QSystemTrayIcon::isSystemTrayAvailable()).toBool();
    APPCONF.keyboardVariant     = settings.value("keyboardVariant", "").toString();
    APPCONF.keyDelay            = settings.value("keyDelay", 2).toInt();
    APPCONF.keyHold             = settings.value("keyHold", 1).toInt();
    APPCONF.typeDelay           = settings.value("typeDelay", 2).toInt();
    APPCONF.typeHold            = settings.value("typeHold", 1).toInt();
    APPCONF.mouseStep           = settings.value("mouseStep", 10).toInt();
    APPCONF.wheelStep           = settings.value("wheelStep", 3).toInt();
    APPCONF.glideStep           = settings.value("glideStep", 2).toInt();
    APPCONF.glideRate           = settings.value("glideRate", 20).toInt();
    APPCONF.numPadGrid          = settings.value("numPadGrid", false).toBool();
    APPCONF.audioId             = settings.value("audioId", QMediaDevices::defaultAudioInput().id()).toByteArray();
    APPCONF.osdOn               = settings.value("osdOn", false).toBool();
    APPCONF.osdAll              = settings.value("osdAll", false).toBool();
    APPCONF.osdDuration         = settings.value("osdDuration", 3000).toInt();
    APPCONF.osdPosition         = settings.value("osdPosition", QPoint(100, 250)).toPoint();
    APPCONF.osdSize             = settings.value("osdSize", QSize(450, 40)).toSize();
    APPCONF.rFontSize           = settings.value("rFontSize", 18).toInt();
    APPCONF.rInvertColor        = settings.value("rInvertColor", false).toBool();
    APPCONF.speechLang          = settings.value("speechLang", APPCONF.lang.replace("_", "-")).toString();
    APPCONF.themeIndex          = settings.value("themeIndex", 0).toInt();
    what->setMouseSteps(APPCONF.mouseStep, APPCONF.glideStep, APPCONF.wheelStep);
    what->setKeyDelays(APPCONF.keyDelay, APPCONF.keyHold, APPCONF.typeDelay, APPCONF.typeHold);
    ui->setTabs->setCurrentIndex(settings.value("mainTabIndex", 0).toInt());    // Restore current index after sets are loaded
}

void MainWindow::writeSettings() {
    QSettings settings(APP_ORG, APP_NAME);
    settings.setValue("appPos",             pos());
    settings.setValue("appSize",            size());
    settings.setValue("appSplit",           ui->splitter->saveState());
    settings.setValue("autoStart",          APPCONF.autoStart);
    settings.setValue("ask2Quit",           APPCONF.ask2Quit);
    settings.setValue("maxlines",           APPCONF.maxlines);
    settings.setValue("startReduced",       APPCONF.startReduced);
    settings.setValue("removeBadUtterance", APPCONF.removeBadUtterance);
    settings.setValue("transparentMode",    APPCONF.transparentMode);
    settings.setValue("osdNoDictation",     APPCONF.osdNoDictation);
    settings.setValue("capitalizeFirst",    APPCONF.capitalizeFirst);
    settings.setValue("dictationSuffix",    APPCONF.dictationSuffix);
    settings.setValue("useSysTray",         APPCONF.useSysTray);
    settings.setValue("keyboardVariant",    APPCONF.keyboardVariant);
    settings.setValue("keyDelay",           APPCONF.keyDelay);
    settings.setValue("keyHold",            APPCONF.keyHold);
    settings.setValue("typeDelay",          APPCONF.typeDelay);
    settings.setValue("typeHold",           APPCONF.typeHold);
    settings.setValue("mouseStep",          APPCONF.mouseStep);
    settings.setValue("wheelStep",          APPCONF.wheelStep);
    settings.setValue("glideStep",          APPCONF.glideStep);
    settings.setValue("glideRate",          APPCONF.glideRate);
    settings.setValue("numPadGrid",         APPCONF.numPadGrid);
    settings.setValue("audioId",            APPCONF.audioId);
    // settings.setValue("audioId",            QMediaDevices::defaultAudioInput().id());
    settings.setValue("osdOn",              APPCONF.osdOn);
    settings.setValue("osdAll",             APPCONF.osdAll);
    settings.setValue("osdDuration",        APPCONF.osdDuration);
    settings.setValue("osdPosition",        APPCONF.osdPosition);
    settings.setValue("osdSize",            APPCONF.osdSize);
    settings.setValue("rFontSize",          APPCONF.rFontSize);
    settings.setValue("rInvertColor",       APPCONF.rInvertColor);
    settings.setValue("mainTabIndex",       ui->setTabs->currentIndex());
    settings.setValue("speechLang",         APPCONF.speechLang);
    settings.setValue("defModel",           APPCONF.model);
    settings.setValue("themeIndex",         APPCONF.themeIndex);
}

void MainWindow::termSignalHandler(int) {   char a = 1; ::write(sigtermFd[0], &a, sizeof(a));   }
void MainWindow::intSignalHandler(int)  {   char a = 1; ::write(sigintFd[0],  &a, sizeof(a));    }
void MainWindow::hupSignalHandler(int)  {   char a = 1; ::write(sighupFd[0],  &a, sizeof(a));    }

void MainWindow::handleSigTerm() {
    snTerm->setEnabled(false);
    char tmp;
    ::read(sigtermFd[1], &tmp, sizeof(tmp));
    std::cerr << APP_STR << " : SIGTERM received, exiting program.\n";
    askForClose();
    snTerm->setEnabled(true);
}

void MainWindow::handleSigInt() {
    snInt->setEnabled(false);
    char tmp;
    ::read(sigintFd[1], &tmp, sizeof(tmp));
    std::cerr << APP_STR << " : SIGINT received, exiting program.\n";
    askForClose();
    snInt->setEnabled(true);
}

void MainWindow::handleSigHup() {
    snHup->setEnabled(false);
    char tmp;
    ::read(sighupFd[1], &tmp, sizeof(tmp));
    std::cerr << APP_STR << " : SIGHUP received, exiting program.\n";
    askForClose();
    snHup->setEnabled(true);
}
