/*
 * SPDX-FileCopyrightText: 2024-2025 Bruno Anselme <be.root@gfree.fr>
 *
 * Authors:
 *   Bruno Anselme <be.root@gfree.fr>
 *
 *   SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This file is part of the NoComprendo software.
*/

#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <QMainWindow>
#include <QFile>
#include <QLabel>
#include <QSocketNotifier>
#include <QComboBox>

#include "about.h"
#include "comprendo.h"
#include "reminder.h"
#include "trayicon.h"
#include "osd.h"
#include "blink.h"
#include "speech.h"
#include "config.h"

QT_BEGIN_NAMESPACE
namespace Ui { class MainWindow; }
QT_END_NAMESPACE

class MainWindow : public QMainWindow {
    Q_OBJECT

public:
    MainWindow(QWidget *parent = nullptr);
    ~MainWindow();
// Unix signals
    static void termSignalHandler(int unused);          //!< Unix signal SIGTERM handler.
    static void intSignalHandler(int unused);           //!< Unix signal SIGINT handler.
    static void hupSignalHandler(int unused);           //!< Unix signal SIGHUP handler.
public Q_SLOTS:
    void    handleSigTerm();                            //!< Qt signal handlers.
    void    handleSigInt();                             //!< Qt signal handlers.
    void    handleSigHup();                             //!< Qt signal handlers.
// End of Unix signals

private Q_SLOTS:
    void    toggleListening(bool);
    void    forceListening(bool);
    void    toggleDictation(bool);                      //!< Start/stop dictation mode
    void    toggleSuspend(bool);                              //!< Suspend/resume listening, update application status
    void    utteranceReceived(QString utterance);       //!< Send utterance to Comprendo object, print utterance, update application
    void    execMeta(Command::metaType m);              //!< Run meta command (connected to Comprendo::metaCommand signal)
    void    openReminder();                             //!< Create Reminder object
    void    closeReminder();                            //!< Delete Reminder object
    void    listeningStatus(bool);                      //!< Change TrayIcon depending on listening and suspend states
    void    raiseWindow();                              //!< Bring MainWindow to front
    void    clearConsole();                             //!< Clear console
    void    newSet();                                   //!< Create a new set
    void    deleteSet();                                //!< Delete selected set
    void    updateCurrentSet(QListWidgetItem*);
    void    information(const QString &mes);            //!< Generic information dialog
    void    askForClose();                              //!< Close confirmation dialog
    void    doSettings();                               //!< Start settings dialog
    void    raiseSetSelector();                         //!< Bring window to front and select lstSets
    void    importSet();
    void    about();
    void    resumeListening();
    void    speak();
    void    renameSet();
    void    copyExamples(bool confirm = true);          //!< Copy defaults examples to user config
    void    showContextMenu(QPoint);
    void    changeAudioSource(int);

protected:
    void    timerEvent(QTimerEvent *) override;

private:
    void    runSysTray();
    void    removeLastOutput();
    void    writeSettings();
    void    readSettings();
    void    updateSets();
    int     lastCommonChar(QString *s1, QString *s2);

private:
    Ui::MainWindow *ui;
    Comprendo      *what;                               // Voice command engine
    bool            firstPartial;
    bool            listening;
    bool            dictation;
    int             dictLength;
    QString         lastDict;
    bool            once;                               // Dictation once (OneDictation)
    QChar           lastPrintedChar;                    // Last non-space char printed (manage lowercase after comma)
    CommandSet::ShortcutFormat  shortcutFormat;         // QtShortcut or XdoShortcut
    Reminder       *reminder;                           // Reminder dialog pointer
    QLabel         *listenState;                        // Right field in status bar (Stopped, Listening, Suspended)
    int             timerid;
    bool            suspended;
    bool            lastOk;
    int             lineCount;
    AboutDialog    *aboutDialog;                        // aboutDialog pointer
    TrayIcon       *tray;
    OSDWindow      *osd;                                // On screen display window
    MouseBlink     *blinker;
    QStringList     multiplierUtterance;
    QString         backSpace;
    Speech          speech;
    bool            speaking;
    QStringList     themes;
    QComboBox      *cbSources;
    // function pointers for meta-commands launching dialogs
    void            (MainWindow::*delayedMethod)();     // Method waiting for execution
    void            (MainWindow::*runningMethod)();     // Method running
    // Signal handler data
    QSocketNotifier    *snTerm, *snInt, *snHup;
};
#endif // MAINWINDOW_H
