/*
 * SPDX-FileCopyrightText: 2024-2025 Bruno Anselme <be.root@gfree.fr>
 *
 * Authors:
 *   Bruno Anselme <be.root@gfree.fr>
 *
 *   SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This file is part of the NoComprendo software.
*/

/**
 * @file      osd.cpp
 * @author    Bruno ANSELME
 * @brief     OSDWindow class definition
 * @details   OSD window. Bypass window manager. Called by OSDEditor.
 */

#include <QScreen>
#include <QPainter>
#include <QShowEvent>
#include <QTimerEvent>
#include "misc.h"
#include "osd.h"

OSDWindow::OSDWindow() :
    QFrame(nullptr, Qt::FramelessWindowHint
           | Qt::WindowStaysOnTopHint
           | Qt::X11BypassWindowManagerHint
           | Qt::WindowDoesNotAcceptFocus
           | Qt::SplashScreen),
    ui(new Ui::OSD),
    message(this),
    resizing(false)
{
    ui->setupUi(this);
    hide();
    rscreens = getDesktopRect();
    changeOsd(APPCONF.osdPosition, APPCONF.osdSize);
    timer.setSingleShot(true);
    timer.setInterval(APPCONF.osdDuration);
    connect(&timer,     SIGNAL(timeout()),  this,   SLOT(endTimer()));
}

void OSDWindow::mousePressEvent(QMouseEvent *event) {
    offset = pos() - event->globalPosition().toPoint();
    resizing = handle.contains(event->pos());
}

void OSDWindow::mouseMoveEvent(QMouseEvent *event)  {
    if (!resizing) {
        changeOsd(event->globalPosition().toPoint() + offset, size());
    } else {
        QSize  newSize(qMax(event->globalPosition().toPoint().x() - pos().x(), 200), qMax(event->globalPosition().toPoint().y() - pos().y(), 40));
        changeOsd(pos(), newSize);
    }
}

void OSDWindow::changeOsd(QPoint pos, QSize siz) {
    setMaximumSize(siz);
    setMinimumSize(siz);
    if (pos.x() > (rscreens.right() - siz.width()))          // Check if OSD window is out of screen
        pos.setX(rscreens.right() - siz.width());
    if (pos.y() > (rscreens.bottom() - siz.height()))
        pos.setY(rscreens.bottom() - siz.height());
    if (pos.x() < 0) pos.setX(0);
    if (pos.y() < 0) pos.setY(0);
    setGeometry(pos.x(), pos.y(), siz.width(), siz.height());
    handle.setRect(rect().left() + rect().width() - 10, rect().top() + rect().height() - 10, 10, 10);
}

void OSDWindow::paintEvent(QPaintEvent *event) {
    QFrame::paintEvent(event);
    if (!timer.isActive()) {                // Editing OSD position and size;
        QPainter    painter(this);
        painter.fillRect(handle, Qt::darkGray);
    }
}

void OSDWindow::setFlashingText(QString s) {
    if (timer.isActive()) timer.stop();
    ui->lbMessage->setStyleSheet("QLabel { }");
    ui->lbMessage->setText(s.simplified());
    show();
    timer.start();
}

void OSDWindow::setItalic(QString s) {
    if (timer.isActive()) timer.stop();
    ui->lbMessage->setStyleSheet("QLabel { font-style: italic; }");
    ui->lbMessage->setText(s.simplified());
    show();
    timer.start();
}

void OSDWindow::setUnknown(QString s) {
    if (timer.isActive()) timer.stop();
    ui->lbMessage->setStyleSheet("QLabel { color: red; }");
    ui->lbMessage->setText(s.simplified());
    show();
    timer.start();
}

void OSDWindow::setText(QString s) {
    ui->lbMessage->setStyleSheet("QLabel { }");
    ui->lbMessage->setText(s.simplified());
    show();
}

void OSDWindow::endTimer() {
    hide();
}
