/*
 * SPDX-FileCopyrightText: 2024-2025 Bruno Anselme <be.root@gfree.fr>
 *
 * Authors:
 *   Bruno Anselme <be.root@gfree.fr>
 *
 *   SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This file is part of the NoComprendo software.
*/

#include <QFile>
#include <QFileDialog>
#include <QMessageBox>
#include <QDrag>
#include <QMimeData>
#include <QMenu>

#include "app_version.h"
#include "settab.h"
#include "commandset.h"
#include "editcommand.h"
#include "ui_settab.h"
#include "misc.h"

static QString clipBoardLine;       //!< clipBoardLine is an internal clipboard, shared between all SetTabs.

SetTab::SetTab(Comprendo *what_, QString &tabName_, QTabWidget *parent) :
    QWidget(parent),
    ui(new Ui::SetTab)
  , tabName(tabName_)
  , comCount(0)
{
    what    = what_;
    newcom   = false;
    ui->setupUi(this);
    cmdFile = QString("%1/%2.noc").arg(APPCONF.userSetsPath.absolutePath()).arg(tabName);

    loadCommandSet();
    ui->twCom->verticalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);

// Drag test
//    ui->twCom->setDragEnabled(true);
//    ui->twCom->setDropIndicatorShown(true);
//    ui->twCom->setDragDropMode(QAbstractItemView::DragOnly);

    connect(ui->btAdd,     SIGNAL(clicked()),                   this,   SLOT(addCommand()));
    connect(ui->btRemove,  SIGNAL(clicked()),                   this,   SLOT(removeCommand()));
    connect(ui->twCom,     SIGNAL(cellDoubleClicked(int,int)),  this,   SLOT(editionCommand(int,int)));
    connect(ui->btModify,  SIGNAL(clicked()),                   this,   SLOT(editionCommand()));
    connect(ui->btExport,  SIGNAL(clicked()),      this,   SLOT(exportSet()));
    connect(ui->twCom, SIGNAL(customContextMenuRequested(const QPoint &)), this, SLOT(ShowContextMenu(const QPoint &)));
    connect(ui->twCom->horizontalHeader(), SIGNAL(sortIndicatorChanged(int, Qt::SortOrder)), this, SLOT(saveCommandSet()));
}

SetTab::~SetTab()
{
    delete ui;
}

void SetTab::ShowContextMenu(const QPoint &pos)
{
    if (ui->twCom->currentRow() == -1)
        return;
    QMenu contextMenu("Context menu", ui->twCom);

    QAction actionCopy(QIcon(":/icons/copy.png"), tr("Copy"), this);
    connect(&actionCopy, SIGNAL(triggered()), this, SLOT(copyCommand()));
    contextMenu.addAction(&actionCopy);

    QAction actionPaste(QIcon(":/icons/paste.png"), tr("Paste"), this);
    connect(&actionPaste, SIGNAL(triggered()), this, SLOT(pasteCommand()));
    contextMenu.addAction(&actionPaste);

    QAction actionDelete(QIcon(":/icons/delete.png"), tr("Delete"), this);
    connect(&actionDelete, SIGNAL(triggered()), this, SLOT(removeCommand()));
    contextMenu.addAction(&actionDelete);

    contextMenu.exec(mapToGlobal(pos));
}

void SetTab::addField(int col, QString s) {
    QTableWidgetItem *newItem = new QTableWidgetItem(s);
    newItem->setFlags(Qt::ItemIsSelectable | Qt::ItemIsEnabled);
// Drag test
//    newItem->setFlags(Qt::ItemIsSelectable | Qt::ItemIsEnabled | Qt::ItemIsDragEnabled );
    ui->twCom->setItem(comCount, col, newItem);
}

QString SetTab::commandToLine(int row) {
    QString line;
    Command::comType ctyp = Command::typeFromName(ui->twCom->item(row, COL_TYPE)->text());
    line.append(QString("%1 : %2").arg(ctyp).arg(ui->twCom->item(row, COL_UTTERANCE)->text()));         // Save command type
    switch (ctyp) {
    case Command::Meta:
        line.append(QString(" : %1").arg(Command::metaFromName(ui->twCom->item(row, COL_COMMAND)->text())));    // Save meta command
        break;
    case Command::Key:
        line.append(QString(" : %1").arg(SEQ_TO_PORTABLE(ui->twCom->item(row, COL_COMMAND)->text())));          // Save command shortcut
        break;
    case Command::Mouse:
        line.append(QString(" : %1").arg(Command::mouseFromName(ui->twCom->item(row, COL_COMMAND)->text())));   // Save mouse command
        break;
    case Command::Text: {
        QString txt(ui->twCom->item(row, COL_COMMAND)->text());
        txt.replace(QString("\n"), QString("\\n"));
        txt.replace(QString("\""), QString("\\\""));
        line.append(" : ").append(Command::addQuotes(txt));                           // Save text input
    }
        break;
    case Command::Exec:
        line.append(" : ").append(ui->twCom->item(row, COL_COMMAND)->text());                           // Save program or script to start
        break;
    }
//    line.append("\n");
    return line;
}

void SetTab::lineToCommand(QString &line) {
    ui->twCom->setSortingEnabled(false);
    ui->twCom->setRowCount(comCount+1);
    QStringList slist = line.split(QString(" : "));
    QString     utt   = slist.at(1);
    addField(COL_TYPE, Command::getTypeName(static_cast<Command::comType>(slist.at(0).toInt())));
    addField(COL_UTTERANCE, utt);
    switch (static_cast<Command::comType>(slist.at(0).toInt())) {
    case Command::Key:
        addField(COL_COMMAND, SEQ_TO_NATIVE(slist.at(2)));
        break;
    case Command::Meta:
        addField(COL_COMMAND, Command::getMetaName(static_cast<Command::metaType>(slist.at(2).toInt())));
        break;
    case Command::Mouse:
        addField(COL_COMMAND, Command::getMouseName(static_cast<Command::mouseType>( slist.at(2).toInt())));
        break;
    case Command::Text: {
        QString txt(Command::removeQuotes(slist.at(2)));
        txt.replace(QString("\\n"), QString("\n"));
        txt.replace(QString("\\\""), QString("\""));
        addField(COL_COMMAND, txt);
    }
        break;
    case Command::Exec:
        addField(COL_COMMAND, slist.at(2));
        break;
    }
    comCount++;
    ui->twCom->setSortingEnabled(true);
}

void SetTab::loadCommandSet() {
    ui->twCom->clearContents();
    ui->twCom->setRowCount(comCount = 0);
    QFile file(cmdFile);
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
        return;
    QTextStream f(&file);
    QString     cmd;
    while (f.readLineInto(&cmd))
        lineToCommand(cmd);
    ui->twCom->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
}

void SetTab::saveCommandSet(bool reload) {
    if (cmdFile.isEmpty())
        return;
    QFile file(cmdFile);
    if (!file.open(QIODevice::WriteOnly | QIODevice::Text))
        return;
    QTextStream f(&file);
    for (int i = 0; i < comCount; i++)
        f <<  commandToLine(i) << "\n";
    file.close();
    if (reload)
        what->loadAllSets();
}

void SetTab::addCommand() {
    ui->twCom->setSortingEnabled(false);
    QString cmd("");
    ui->twCom->setRowCount(comCount+1);
    addField(COL_TYPE, Command::getTypeName(Command::Key));
    addField(COL_UTTERANCE, cmd.trimmed());
    addField(COL_COMMAND, "");
    ui->twCom->setCurrentCell(comCount,0);
    ui->twCom->scrollToBottom();
    comCount++;
    ui->twCom->selectRow(comCount-1);
    newcom = true;
    ui->twCom->setSortingEnabled(true);
    emit ui->twCom->cellDoubleClicked(ui->twCom->currentRow(),0);   // Enter editionCommand
}

void SetTab::copyCommand() {
    clipBoardLine = commandToLine(ui->twCom->currentRow());
}

void SetTab::pasteCommand() {
    if (!clipBoardLine.isEmpty()) {
//        qWarning() << clipBoardLine;
        lineToCommand(clipBoardLine);
        saveCommandSet(true);
    }
}

void SetTab::removeCommand() {
    if (ui->twCom->currentRow() != -1) {
        ui->twCom->removeRow(ui->twCom->currentRow());
        comCount--;
        ui->twCom->resizeColumnsToContents();
        saveCommandSet(true);
    }
}

void SetTab::editionCommand() {
    if (ui->twCom->currentRow() != -1)
        editionCommand(ui->twCom->currentRow(),0);
}

void SetTab::editionCommand(int row, int) {
    Command *cmd = nullptr;
    switch (Command::typeFromName(ui->twCom->item(row, COL_TYPE)->text())) {
    case Command::Meta:
        cmd = Command::createMetaCommand(Command::metaFromName(ui->twCom->item(row, COL_COMMAND)->text()));
        break;
    case Command::Key:
        cmd = Command::createKeyCommand(SEQ_TO_PORTABLE(ui->twCom->item(row, COL_COMMAND)->text()));
        break;
    case Command::Mouse:
        cmd = Command::createMouseCommand(Command::mouseFromName(ui->twCom->item(row, COL_COMMAND)->text()));
        break;
    case Command::Text:
        cmd = Command::createTextCommand(ui->twCom->item(row, COL_COMMAND)->text());
        break;
    case Command::Exec:
        cmd = Command::createExecCommand(ui->twCom->item(row, COL_COMMAND)->text());
        break;
    }

    bool memListen = what->isListening();
    emit setListening(false);
    EditCommand dial(ui->twCom->item(row, COL_UTTERANCE)->text(), *cmd, what);
    int dialogCode = dial.exec();

    if (dialogCode == QDialog::Accepted) {
        Command::comType ctyp = dial.getType();
        ui->twCom->item(row, COL_UTTERANCE)->setText(dial.getUtterance());
        ui->twCom->item(row, COL_TYPE)->setText(Command::getTypeName(ctyp));
        switch (ctyp) {
        case Command::Meta:
            ui->twCom->item(row, COL_COMMAND)->setText(Command::getMetaName(dial.getMeta()));
            break;
        case Command::Key:
            ui->twCom->item(row, COL_COMMAND)->setText(SEQ_TO_NATIVE(dial.getSequence().toString()));
            break;
        case Command::Mouse:
            ui->twCom->item(row, COL_COMMAND)->setText(Command::getMouseName(dial.getMouse()));
            break;
        case Command::Text: {
            ui->twCom->item(row, COL_COMMAND)->setText(dial.getText());
        }
            break;
        case Command::Exec:
            ui->twCom->item(row, COL_COMMAND)->setText(dial.getExec());
            break;
        }
        ui->twCom->resizeColumnsToContents();
        saveCommandSet(true);
    } else if (dialogCode == QDialog::Rejected) {
        if (newcom) {
            removeCommand();
        }
    }
    newcom = false;
    if (cmd) delete cmd;
    emit setListening(memListen);
    ui->twCom->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
}

void SetTab::exportSet() {
    QString fileName = QFileDialog::getSaveFileName(this, tr("Save commands set"),
                                                    QString("%1/%2.noc").arg(QDir::homePath()).arg(tabName),
                                                    QString("%1 (*.noc)").arg(APP_NAME));
    if (!fileName.isEmpty()) {
        QFile file(fileName);
        if (!file.open(QIODevice::WriteOnly | QIODevice::Text))
            return;
        QTextStream f(&file);
        f << readFile(cmdFile);
        file.close();
        QMessageBox msgBox;
        msgBox.setText(tr("Exported in file : <b>%1</b>").arg(fileName));
        msgBox.exec();
    }
}

void SetTab::changeEvent(QEvent *e) {
    QWidget::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}
