/*
 * SPDX-FileCopyrightText: 2024-2025 Bruno Anselme <be.root@gfree.fr>
 *
 * Authors:
 *   Bruno Anselme <be.root@gfree.fr>
 *
 *   SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This file is part of the NoComprendo software.
*/

/**
 * @file      settings.cpp
 * @author    Bruno ANSELME
 * @brief     Settings class definition
 * @details   Settings dialog for NoComprendo.
 */

#include <QDebug>
#include <iostream>
#include <QSystemTrayIcon>
#include <QSettings>

#include "app_version.h"
#include "settings.h"
#include "misc.h"
#include "osdeditor.h"
#include "modelchooser.h"
#include "zoombutton.h"
#include "editutterance.h"
#include "modelloader.h"

Settings::Settings(Comprendo *what_) : QDialog(),
    ui(new Ui::NCSettings)
{
    what = what_;

    ui->setupUi(this);
    QLocale     loc(APPCONF.lang);
    ui->ckAutostart->setChecked(APPCONF.autoStart);
    ui->ckAsk2Quit->setChecked(APPCONF.ask2Quit);
    ui->sprMaxLines->setValue(APPCONF.maxlines);
    ui->ckStartReduced->setChecked(APPCONF.startReduced);
    ui->ckRemoveBadUtterance->setChecked(APPCONF.removeBadUtterance);
    ui->cbTransparent->setCurrentIndex(APPCONF.transparentMode ? 0 : 1);
    ui->ckOsdNoDictation->setChecked(APPCONF.osdNoDictation);
    ui->ckCapitalize->setChecked(APPCONF.capitalizeFirst);
    ui->teDictSuffix->setText(APPCONF.dictationSuffix);
    ui->ckSysTray->setChecked(APPCONF.useSysTray);
    ui->ckSysTray->setEnabled(QSystemTrayIcon::isSystemTrayAvailable());
    ui->lbLangValue->setText(QString("<b>%3</b> (%1/%2)")
        .arg(loc.nativeLanguageName())
        .arg(loc.nativeTerritoryName())
        .arg(loc.name()));
    ui->lbModelValue->setText(QString("<b>%1</b>").arg(APPCONF.model));
    ui->ckOsdOn->setChecked(APPCONF.osdOn);
    ui->ckOsdAll->setChecked(APPCONF.osdAll);
    ui->sbOsdDuration->setValue(APPCONF.osdDuration/1000);
    ui->sbKeyDelay->setValue(APPCONF.keyDelay);
    ui->sbKeyHold->setValue(APPCONF.keyHold);
    ui->sbTypeDelay->setValue(APPCONF.typeDelay);
    ui->sbTypeHold->setValue(APPCONF.typeHold);
    ui->spMouseStep->setValue(APPCONF.mouseStep);
    ui->spWheelStep->setValue(APPCONF.wheelStep);
    ui->spGlideStep->setValue(APPCONF.glideStep);
    ui->spGlideRate->setValue(APPCONF.glideRate);
    ui->ckNumpadGrid->setChecked(APPCONF.numPadGrid);
    osdPosition = APPCONF.osdPosition;
    osdSize     = APPCONF.osdSize;
    ui->sprFontSize->setValue(APPCONF.rFontSize);
    ui->ckInvertColor->setChecked(APPCONF.rInvertColor);
    ui->cbTheme->setCurrentIndex(APPCONF.themeIndex);

    updateModelList();

    // Fill UI with mouse moveto commands
    numPadMap = { { 1, 7 }, { 2, 8 }, { 3, 9 }, { 4, 4 }, { 5, 5 }, { 6, 6 }, { 7, 1 }, { 8, 2 }, { 9, 3 }, };
    QStringList box = ZoomButton::readMouseFileList();
    ui->tbMouseCom->setRowCount(box.count());
    ui->tbMouseCom->verticalHeader()->hide();
    for (int i = 0 ; i < box.count(); i++ ) {
        QString name = i ? tr("Cell %1").arg(i) : tr("Backward");
        QTableWidgetItem *newItem = new QTableWidgetItem(name);
        newItem->setTextAlignment(Qt::AlignRight | Qt::AlignVCenter);
        ui->tbMouseCom->setItem(i, 0, newItem);
        newItem = new QTableWidgetItem(box[i]);
        ui->tbMouseCom->setItem(i, 1, newItem);
    }
    ui->tbMouseCom->resizeRowsToContents();
    initGrid();

    // Fill combo keyboard variants
    QString langShort =  loc.name().left(2);
    QString program = findProgram("localectl");
    QProcess aProcess;
    aProcess.start(program, { "list-x11-keymap-variants", langShort});
    aProcess.waitForFinished();
    QString output = aProcess.readAllStandardOutput();
    QStringList variants = output.split("\n", Qt::SkipEmptyParts);
    ui->cbKeyboardVariant->addItems(variants);
    ui->cbKeyboardVariant->setCurrentIndex(variants.indexOf(APPCONF.keyboardVariant) + 1);

    // Fill UI with repetition commands
    QStringList multiplierUtterance
            = { "once", "twice", "three times", "four times", "five times", "six times", "seven times", "eight times", "nine times", "ten times"};
    QString txt =   readFile(APPCONF.repeaterFile).trimmed();
    if (!txt.isEmpty())
        multiplierUtterance = txt.split("\n", Qt::SkipEmptyParts);
    ui->tbRepetition->setRowCount(multiplierUtterance.count());
    ui->tbRepetition->verticalHeader()->hide();
    for (int j = 0; j < multiplierUtterance.count() ; j++) {
        QTableWidgetItem *newItem = new QTableWidgetItem(QString("× %1").arg(j + 1));
        newItem->setTextAlignment(Qt::AlignHCenter | Qt::AlignVCenter);
        ui->tbRepetition->setItem(j, 0, newItem);
        newItem = new QTableWidgetItem(multiplierUtterance[j]);
        ui->tbRepetition->setItem(j, 1, newItem);
    }
    ui->tbRepetition->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    ui->tbRepetition->resizeRowsToContents();
    ui->btRemoveMult->setEnabled(ui->tbRepetition->rowCount() > 10);

    // Fill combo speech languages
    int n = 0;
    int goodLine = 0;
    QMapIterator<QString, QString> m(APPCONF.speechesList);
    while (m.hasNext()) {
        m.next();
        ui->cbSpeechLang->addItem(m.value(), m.key());
        if (m.key() == APPCONF.speechLang)
            goodLine = n;
        n++;
    }
    ui->cbSpeechLang->setCurrentIndex(goodLine);

    QSettings settings(APP_ORG, APP_NAME);
    move(settings.value("confPos", QPoint(100, 100)).toPoint());
    resize(settings.value("confSize", QSize(454, 519)).toSize());
    ui->tabSettings->setCurrentIndex(settings.value("settingsTabIndex", 0).toInt());
    ui->tbMouseCom->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);

    connect(ui->btOsdPosition,  SIGNAL(clicked()),                      this,   SLOT(moveOsd()));
    connect(ui->ckNumpadGrid,   SIGNAL(clicked(bool)),                  this,   SLOT(numpadGridChanged(bool)));
    connect(ui->tbMouseCom,     SIGNAL(currentCellChanged(int, int, int, int)),
            this,               SLOT(changeMouseLine(int, int, int, int)));
    connect(ui->tbMouseCom,     SIGNAL(cellDoubleClicked(int, int)),    this,   SLOT(editMouseUtterance(int, int)));
    connect(ui->tbRepetition,   SIGNAL(cellDoubleClicked(int, int)),    this,   SLOT(editRepeaterUtterance(int, int)));
    connect(ui->btLoadModel,    SIGNAL(clicked(bool)),                  this,   SLOT(loadModel()));
    connect(ui->btAddMult,      SIGNAL(clicked()),                      this,   SLOT(addMultiplier()));
    connect(ui->btRemoveMult,   SIGNAL(clicked()),                      this,   SLOT(removeLastMultiplier()));
    connect(ui->btResetSets,    SIGNAL(clicked()),                      this,   SLOT(sendResetSets()));
    connect(ui->cbSpeechLang,   SIGNAL(currentIndexChanged(int)),       this,   SLOT(speechLangChanged(int)));
    connect(ui->btRestoreDelays,SIGNAL(clicked()),                      this,   SLOT(restoreDelays()));
}

Settings::~Settings() {
    QSettings settings(APP_ORG, APP_NAME);
    settings.setValue("confPos",    pos());
    settings.setValue("confSize",   size());
    settings.setValue("settingsTabIndex",   ui->tabSettings->currentIndex());
}

void Settings::initGrid() {
    QLayoutItem *child;
    while ((child = ui->gridCells->takeAt(ui->gridCells->count() - 1)) != nullptr) {
        delete child->widget(); // delete the widget
        delete child;           // delete the layout item
    }
    QFont myFont;
    myFont.setBold(true);
    for (int i = 0; i < 9; i++) {
        QLabel *cell = new QLabel(ui->tbMouseCom);
        cell->setMinimumSize(QSize(0, 25));
        cell->setAutoFillBackground(true);
        cell->setFrameShape(QFrame::StyledPanel);
        cell->setFrameShadow(QFrame::Raised);
        cell->setText(QString::number(i + 1));
        cell->setFont(myFont);
        cell->setAlignment(Qt::AlignCenter);
        if (APPCONF.numPadGrid)
            ui->gridCells->addWidget(cell, 2 - (i / 3), i % 3, 1, 1);
        else
            ui->gridCells->addWidget(cell, i / 3, i % 3, 1, 1);
    }
    ui->tbMouseCom->setCurrentCell(-1, -1);
}

void Settings::numpadGridChanged(bool ck) {
    APPCONF.numPadGrid = ck;
    initGrid();
}

void Settings::speechLangChanged(int) {
    APPCONF.speechLang = ui->cbSpeechLang->currentData().toString();
}

void Settings::sendResetSets() {
    emit resetDefaultSets();
}

void Settings::restoreDelays() {
    ui->sbKeyDelay->setValue(2);
    ui->sbKeyHold->setValue(1);
    ui->sbTypeDelay->setValue(2);
    ui->sbTypeHold->setValue(1);
}

void Settings::addMultiplier() {
    ui->tbRepetition->setRowCount(ui->tbRepetition->rowCount() + 1);
    int j = ui->tbRepetition->rowCount() - 1;
    QTableWidgetItem *newItem = new QTableWidgetItem(QString("× %1").arg(j + 1));
    newItem->setTextAlignment(Qt::AlignRight | Qt::AlignVCenter);
    newItem->setTextAlignment(Qt::AlignHCenter | Qt::AlignVCenter);
    ui->tbRepetition->setItem(j, 0, newItem);
    newItem = new QTableWidgetItem(tr("Record something"));
    ui->tbRepetition->setItem(j, 1, newItem);
    editRepeaterUtterance(j, 1);
    ui->btRemoveMult->setEnabled(ui->tbRepetition->rowCount() > 10);
}

void Settings::removeLastMultiplier() {
    ui->tbRepetition->removeRow(ui->tbRepetition->rowCount() - 1);
    ui->btRemoveMult->setEnabled(ui->tbRepetition->rowCount() > 10);
}

void Settings::updateModelList() {
    ui->cbModel->clear();
    QStringList filter = { "vosk-model-*" };
    QStringList files = APPCONF.voskSystemPath.entryList(filter, QDir::Dirs);   // Load availables models in system path
    int i = 0;
    for (; i < files.count(); i++) {
        ui->cbModel->addItem(QString("%1").arg(files[i]), APPCONF.voskSystemPath.path());
    }
    files += APPCONF.voskUserPath.entryList(filter);                // Add availables models in user path
    for (;i < files.count(); i++) {
        ui->cbModel->addItem(QString("%1").arg(files[i]), APPCONF.voskUserPath.path());
    }
    for (int j=0; j < files.count(); j++) {                         // auto select default language if present
        if (ui->cbModel->itemData(j).toString() + "/" + ui->cbModel->itemText(j) == APPCONF.model)
            ui->cbModel->setCurrentIndex(j);
    }
}

void Settings::loadModel() {
    ModelLoader dial;
    dial.exec();
    updateModelList();
}

void Settings::changeMouseLine(int currentRow, int, int previousRow, int ) {
    if (previousRow > 0) {
        if (APPCONF.numPadGrid)
            previousRow = numPadMap[previousRow];
        previousRow--;
        ui->gridCells->itemAtPosition(previousRow / 3, previousRow % 3)->widget()->setStyleSheet("");
    }
    if (currentRow > 0) {
        if (APPCONF.numPadGrid)
            currentRow = numPadMap[currentRow];
        currentRow--;
        ui->gridCells->itemAtPosition(currentRow / 3, currentRow % 3)->widget()->setStyleSheet("background-color: rgb(180, 180, 190);");
    }
}

void Settings::editMouseUtterance(int row, int ) {
    EditUtterance dial(QString("%1").arg(row ? tr("cell %1").arg(row) : tr("backward")), ui->tbMouseCom->item(row, 1)->text(), what);
    int dialogCode = dial.exec();
    if (dialogCode == QDialog::Accepted) {
        ui->tbMouseCom->item(row, 1)->setText(dial.getUtterance());
    }
}

void Settings::editRepeaterUtterance(int row, int ) {
    EditUtterance dial(QString("%1").arg(row + 1), ui->tbRepetition->item(row, 1)->text(), what);
    int dialogCode = dial.exec();
    if (dialogCode == QDialog::Accepted) {
        ui->tbRepetition->item(row, 1)->setText(dial.getUtterance());
        ui->tbRepetition->resizeRowsToContents();
    }
}

void Settings::accept() {
    // Write mouse file
    QString out;
    for (int i = 0 ; i < ui->tbMouseCom->rowCount(); i++)
        out.append(QString("%1\n").arg(ui->tbMouseCom->item(i, 1)->text()));
    writeFile(APPCONF.mouseFile, out);

    // Write repeater file
    out = "";
    for (int i = 0 ; i < ui->tbRepetition->rowCount(); i++)
        out.append(QString("%1\n").arg(ui->tbRepetition->item(i, 1)->text()));
    writeFile(APPCONF.repeaterFile, out);

    // Write language model
    QString model = ui->cbModel->currentData().toString() + "/" + ui->cbModel->currentText();
    if (APPCONF.model != model) {
        APPCONF.model = model;
        emit sendInformation(tr("Language model change will take effect when the application restart."));
    }

    QString memoKeyboard = QString(APPCONF.keyboardVariant.data());

    // Save information from dialog into singleton
    APPCONF.autoStart           = ui->ckAutostart->isChecked();
    APPCONF.ask2Quit            = ui->ckAsk2Quit->isChecked();
    APPCONF.maxlines            = ui->sprMaxLines->value();
    APPCONF.startReduced        = ui->ckStartReduced->isChecked();
    APPCONF.removeBadUtterance  = ui->ckRemoveBadUtterance->isChecked();
    APPCONF.osdNoDictation      = ui->ckOsdNoDictation->isChecked();
    APPCONF.capitalizeFirst     = ui->ckCapitalize->isChecked();
    APPCONF.dictationSuffix     = ui->teDictSuffix->toPlainText();
    APPCONF.transparentMode     = ui->cbTransparent->currentIndex() == 0;
    APPCONF.useSysTray          = ui->ckSysTray->isChecked();
    APPCONF.mouseStep           = ui->spMouseStep->value();
    APPCONF.glideStep           = ui->spGlideStep->value();
    APPCONF.wheelStep           = ui->spWheelStep->value();
    APPCONF.glideRate           = ui->spGlideRate->value();
    APPCONF.numPadGrid          = ui->ckNumpadGrid->isChecked();
    APPCONF.osdOn               = ui->ckOsdOn->isChecked();
    APPCONF.osdAll              = ui->ckOsdAll->isChecked();
    APPCONF.osdDuration         = ui->sbOsdDuration->value()*1000;
    APPCONF.osdPosition         = osdPosition;
    APPCONF.osdSize             = osdSize;
    APPCONF.keyboardVariant     = (ui->cbKeyboardVariant->currentIndex() != 0) ? ui->cbKeyboardVariant->currentText() : "";
    APPCONF.keyDelay            = ui->sbKeyDelay->value();
    APPCONF.keyHold             = ui->sbKeyHold->value();
    APPCONF.typeDelay           = ui->sbTypeDelay->value();
    APPCONF.typeHold            = ui->sbTypeHold->value();
    APPCONF.rFontSize           = ui->sprFontSize->value();
    APPCONF.rInvertColor        = ui->ckInvertColor->isChecked();
    APPCONF.themeIndex          = ui->cbTheme->currentIndex();

    if (memoKeyboard != APPCONF.keyboardVariant) {
        what->restartDotool();
    }

    QDialog::accept();
}

void Settings::moveOsd() {
    OSDEditor   osde;
    osde.exec();
    osdPosition = osde.getPosition();
    osdSize     = osde.getSize();
}
