/*
 * SPDX-FileCopyrightText: 2024-2025 Bruno Anselme <be.root@gfree.fr>
 *
 * Authors:
 *   Bruno Anselme <be.root@gfree.fr>
 *
 *   SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This file is part of the NoComprendo software.
*/

/**
 * @file      zoombutton.cpp
 * @author    Bruno ANSELME
 * @brief     ZoomButton class definition
 * @details   Split window into 9 boxes, 4 times to get a mouse location.
 *              Special button class for Mouse MoveTo command. Based on keypads numbers.
 */

#include <iostream>
#include <QGuiApplication>
#include <QPainter>
#include <QEventLoop>
#include <QKeyEvent>
#include "zoombutton.h"
#include "misc.h"

static QStringList  boxCommands = { "backward", "one", "two", "three", "four", "five", "six", "seven", "eight", "nine" };

ZoomButton::ZoomButton(Comprendo *comp) : QAbstractButton()
  , done(false)
  , here(0,0)
  , offset(0,0)
{
    setWindowFlags(Qt::FramelessWindowHint
                   | Qt::WindowStaysOnTopHint
                   | Qt::BypassWindowManagerHint);
    screen = getDesktopRect();
    setGeometry(screen.toRect());
    setWindowOpacity(0.6);
    frames.append(rect());
    installEventFilter(this);
    what    = comp;
    timerid = startTimer(250);
    boxKeys = { Qt::Key_Escape,  Qt::Key_1, Qt::Key_2, Qt::Key_3, Qt::Key_4, Qt::Key_5, Qt::Key_6, Qt::Key_7, Qt::Key_8, Qt::Key_9 };
    readMouseFile();
    show();
    activateWindow();       // Required to activate keyboard input with BypassWindowManagerHint
}

ZoomButton::~ZoomButton() {
    what->resetVosk();
}

void ZoomButton::createMouseFile() {
    QString out;
    for (int i = 0 ; i < boxCommands.count(); i++) {
        out.append(QString("%1\n").arg(boxCommands[i]));
    }
    writeFile(APPCONF.mouseFile, out);
}

void ZoomButton::readMouseFile() {
    if (!QFile::exists(APPCONF.mouseFile))      // Create file if not existing
        createMouseFile();
    QString txt =   readFile(APPCONF.mouseFile);
    QStringList     mouseCmds;
    mouseCmds = txt.trimmed().split("\n");
    for (int i = 0; i < mouseCmds.count(); i++) {
        boxCommands[i] = mouseCmds[i];
        boxes[mouseCmds[i]] = boxKeys[i];
    }
}

QStringList ZoomButton::readMouseFileList() {
    if (!QFile::exists(APPCONF.mouseFile))      // Create file if not existing
        createMouseFile();
    QStringList box;
    QString txt =   readFile(APPCONF.mouseFile);
    box = txt.trimmed().split("\n");
    return box;
}

QPointF ZoomButton::whereRatio() {
    return QPointF((qreal) here.x() / (qreal) screen.width()
                 , (qreal) here.y() / (qreal) screen.height());
}

void ZoomButton::exec() {
    QEventLoop wait_loop;
    connect(this,   SIGNAL(clicked()),             &wait_loop, SLOT(quit()));
    connect(this,   SIGNAL(focusLost(bool)),       &wait_loop, SLOT(quit()));
    connect(this,   SIGNAL(zoomClicked()),         &wait_loop, SLOT(quit()));
    wait_loop.exec();   // Wait for a click or focus lost or mouse move finished
 }

bool ZoomButton::eventFilter(QObject *, QEvent *event) {
   if (event->type() == QEvent::WindowDeactivate)
      emit focusLost(true);
   return false;
}

void ZoomButton::keyReleaseEvent(QKeyEvent *e) {
    Qt::Key code = (Qt::Key) e->key();
    if (code == Qt::Key_Escape)                         popFrame();
    if ((code >= Qt::Key_1) && (code <= Qt::Key_9))     pushFrame(code);
}

void ZoomButton::timerEvent(QTimerEvent *t) {
    if (t->timerId()==timerid) {
        QString *utt = what->recognize_from_microphone();

        Comprendo::VoskStatus status = what->getStatus();
        switch(status) {
        case Comprendo::VoskStatus::Partial:
            break;
        case Comprendo::VoskStatus::Final:
        {
            QStringList words = utt->split(" ");
            for (int i = 0; i < words.count(); i++) {
                if (boxes.contains(words[i])) {
                    pushFrame(boxes[words[i]]);
                }
            }
        }
            what->resetVosk();
            break;
        case Comprendo::VoskStatus::Empty:
            break;
        }
        delete utt;        // Utterance created in recognize_from_microphone must be deleted
    }
}

void ZoomButton::pushFrame(Qt::Key key) {
    if (key == Qt::Key_Escape) {         // Box 0 means backward
        popFrame();
        return;
    }
    int num = (int) key - Qt::Key_0;
    QRect last = frames.last();
    QRect r = rectNumber(last, num);
    if (frames.count()<4) {
        frames.append(r);
        update();
    } else {
        here = QPoint(r.center())+offset;
        done = true;
        emit zoomClicked();
    }
}

void ZoomButton::popFrame() {
    frames.removeLast();
    if (frames.isEmpty())
        emit clicked();
    else update();
}

void ZoomButton::resizeEvent(QResizeEvent *e) {
    frames.clear();
    frames.append(QRect(QPoint(0,0) ,e->size()));
}

const QRect ZoomButton::rectNumber(QRect r, int num) {
    int w = r.width()/3;
    int h = r.height()/3;
    num--;
    if (APPCONF.numPadGrid)
        return QRect((num % 3)*w + r.x(), ((8 - num) / 3)*h + r.y(), w, h);     // Keypad order
    else
        return QRect((num % 3)*w + r.x(), (num / 3)*h + r.y(), w, h);           // Natural order, left to right, top to bottom
}

void ZoomButton::drawGrid(QRect r, QPainter *painter, int level) {
    int h = r.height()/3;
    QFont font = painter->font();
    font.setPixelSize(2*h/3);
    painter->setFont(font);
    for(int count=1 ; count<10; count++)
        painter->drawRect(rectNumber(r, count));
    if (level == (frames.count()-1))
        for(int count=1 ; count<10; count++)
            painter->drawText(rectNumber(r, count), Qt::AlignHCenter|Qt::AlignVCenter, QString("%1").arg(count));
}

void ZoomButton::paintEvent(QPaintEvent *) {
    QPainter    painter(this);
    painter.setPen(Qt::black);
    painter.fillRect(rect(), Qt::darkGray);
    for(int i = 0; i < frames.count(); i++) {
        drawGrid(frames[i], &painter, i);
    }
}
